import numpy as np
from menpo.shape import PointCloud, TriMesh, ColouredTriMesh, TexturedTriMesh
from .vtkutils import trimesh_to_vtk, VTKClosestPointLocator


def barycentric_coordinates(point, a, b, c):
    T = np.hstack(((a - c)[:, None], (b - c)[:, None]))
    return np.linalg.lstsq(T, point - c)[0]


def barycentric_coordinates_for_indices(mesh, tri_index, point):
    a, b, c = mesh.points[mesh.trilist[tri_index]]
    return barycentric_coordinates(point, a, b, c)


def barycentric_points_from_contained_points(self, pointcloud, tri_index):
    # http://gamedev.stackexchange.com/questions/23743/whats-the-most-efficient-way-to-find-barycentric-coordinates
    abc_per_tri = self.points[self.trilist[tri_index]]

    a = abc_per_tri[:, 0, :]
    b = abc_per_tri[:, 1, :]
    c = abc_per_tri[:, 2, :]
    p = pointcloud.points

    v0 = b - a
    v1 = c - a
    v2 = p - a
    d00 = (v0 * v0).sum(axis=1)
    d01 = (v0 * v1).sum(axis=1)
    d11 = (v1 * v1).sum(axis=1)
    d20 = (v2 * v0).sum(axis=1)
    d21 = (v2 * v1).sum(axis=1)
    denom = d00 * d11 - d01 * d01
    v = (d11 * d20 - d01 * d21) / denom
    w = (d00 * d21 - d01 * d20) / denom
    u = 1 - v - w

    return np.vstack([u, v, w]).T


def snap_pointcloud_to_surface(self, pointcloud):
    r"""Constrain points in a :map:`PointCloud: to lie as close as possible to
    this :map:`TriMesh`.

    Each point in the provided `pointcloud` will be projected to the surface
    of this mesh in the minimum possible distance.

    Parameters
    ----------
    pointcloud : :map:`PointCloud`
        The pointcloud that will be projected onto this mesh.

    Returns
    -------
    snapped_pointcloud, tri_indices : :map:`PointCloud`, ``(n_points, )`` `ndarray`
        A tuple of the snapped :map`PointCloud` and the indices of the
        triangles involved in the snapping.
    """
    vtk_mesh = trimesh_to_vtk(self)
    locator = VTKClosestPointLocator(vtk_mesh)
    snapped_points, indices = locator(pointcloud.points)
    return PointCloud(snapped_points, copy=False), indices


def barycentric_coordinates_of_pointcloud(self, pointcloud):
    r"""Return the barycentric coordinates of points in a :map:`PointCloud:
    that have been constrained to lie as close as possible to this
    :map:`TriMesh`.

    Each point in the provided `pointcloud` will be projected to the surface
    of this mesh in the minimum possible distance. The barycentric coordinates
    and the triangle index of each point will be returned.

    Parameters
    ----------
    pointcloud : :map:`PointCloud`
        The pointcloud that will be projected onto this mesh.

    Returns
    -------
    bcoords, tri_indices : ``(n_points, 3)`` `ndarray`, ``(n_points, )`` `ndarray`
        A tuple of the barycentric coordinates and the indices of the
        triangles involved for each snapped point.
    """
    p, i = self.snap_pointcloud_to_surface(pointcloud)
    bc = barycentric_points_from_contained_points(self, p, i)
    return bc, i


def barycentric_coordinate_interpolation(
    self, per_vertex_interpolant, bcoords, tri_indices
):
    r"""Interpolate some per-vertex value on this mesh using barycentric coordinates.

    Parameters
    ----------
    per_vertex_interpolant : ``(n_points, k)`` `ndarray`
        Any array of per-vertex data. This will be linearly blended over every
         triangle referenced in ``tri_indices`` with the relevent weights
         given in ``bcoords``.
    bcoords : ``(n_samples, 3)`` `ndarray`
        The barycentric coordinates that will be used in the projection
    tri_indices : ``(n_samples, )`` `ndarray`
        The index of the triangle that the above ``bcoords`` correspond to.

    Returns
    -------
    `ndarray` : ``(n_samples, k)``
        The interpolated values of ``per_vertex_interpolant``.
    """
    shape = per_vertex_interpolant.shape
    if not len(shape) == 2 or shape[0] != self.n_points:
        raise ValueError("per_vertex_interpolant must be of shape (n_points, k)")
    t = per_vertex_interpolant[self.trilist[tri_indices]]
    return np.sum(t * bcoords[..., None], axis=1)


def project_barycentric_coordinates(self, bcoords, tri_indices):
    r"""Projects a set of barycentric coordinates onto this mesh surface,
    returning a :map:`PointCloud`.

    Parameters
    ----------
    bcoords: ``(n_samples, 3)`` `ndarray`
        The barycentric coordinates that will be used in the projection
    tri_indices : ``(n_samples, )`` `ndarray`
        The index of the triangle that the above ``bcoords`` correspond to.

    Returns
    -------
    pointcloud : :map:`PointCloud`
        A :map:`PointCloud` representation of the provided barycentric
        coordinates on this :map:`TriMesh`
    """
    # Interpolate self.points using the more generic
    # barycentric_coordinate_interpolation method
    interped_points = self.barycentric_coordinate_interpolation(
        self.points, bcoords, tri_indices
    )
    return PointCloud(interped_points, copy=False)


def sample_texture_with_barycentric_coordinates_colour(self, bcoords, tri_indices):
    return self.barycentric_coordinate_interpolation(self.colours, bcoords, tri_indices)


def sample_texture_with_barycentric_coordinates_texture(
    self, bcoords, tri_indices, order=1
):
    sample_points = self.barycentric_coordinate_interpolation(
        self.tcoords_pixel_scaled().points, bcoords, tri_indices
    )
    texture = self.texture
    if hasattr(texture, "as_unmasked"):
        # TODO this as_unmasked should not be needed, but it is (we can fall
        # off the texture at bcoords). This means the sampled texture contains
        # wrong (black) values.
        texture = texture.as_unmasked(copy=False)
    return texture.sample(sample_points, order=order).T


TriMesh.snap_pointcloud_to_surface = snap_pointcloud_to_surface
TriMesh.barycentric_coordinates_of_pointcloud = barycentric_coordinates_of_pointcloud
TriMesh.barycentric_coordinate_interpolation = barycentric_coordinate_interpolation
TriMesh.project_barycentric_coordinates = project_barycentric_coordinates
ColouredTriMesh.sample_texture_with_barycentric_coordinates = (
    sample_texture_with_barycentric_coordinates_colour
)
TexturedTriMesh.sample_texture_with_barycentric_coordinates = (
    sample_texture_with_barycentric_coordinates_texture
)



try:
    from .opengl import GLRasterizer
except ImportError:
    pass
from .cpu import (
    rasterize_barycentric_coordinates,
    rasterize_barycentric_coordinate_images,
)
from .base import (
    rasterize_mesh_from_barycentric_coordinate_images,
    rasterize_shape_image_from_barycentric_coordinate_images,
    rasterize_mesh,
)
from .transform import model_to_clip_transform, clip_to_image_transform



import numpy as np

from menpo.base import name_of_callable, Copyable
from menpo.shape import ColouredTriMesh, TexturedTriMesh, PointCloud


class MorphableModel(Copyable):
    r"""
    Abstract class for defining a Morphable Model.

    There are two subclasses, analogously to the :map:`TexturedTriMesh` and
    :map:`ColouredTriMesh`. The only differences come in two areas:

        1. How the shape and texture PCA models are combined in synthesising a
           new instance
        2. How the texture is sampled at locations in forming the cost
           function

    The vast majority of the code is shared in this superclass.

    Please see the references for a basic list of relevant papers.

    Parameters
    ----------
    shape_model : `menpo.model.PCAModel`
        The PCA model of the 3D shape. It is assumed that a shape instance is
        defined as a `menpo.shape.TriMesh`.
    texture_model : `menpo.model.PCAVectorModel`
        The PCA texture model. Note that the texture model can be feature
        based, in which case you need to specify the employed
        `holistic_features`.
    landmarks : `menpo.shape.PointUndirectedGraph`
        The set of sparse landmarks defined in the 3D space.
    holistic_features : `function`
        The features that were used for building the texture model. Please
        refer to `menpo.feature` for a list of potential features.
    diagonal : `int`
        This parameter was used to rescale the training images so that the
        diagonal of their bounding boxes matches the provided value. In other
        words, this parameter defined the size of the training images before
        extracting features.

    References
    ----------
    .. [1] V. Blanz, T. Vetter. "A morphable model for the synthesis of 3D
        faces", Conference on Computer Graphics and Interactive Techniques,
        pp. 187-194, 1999.
    .. [2] P. Paysan, R. Knothe, B. Amberg, S. Romdhani, T. Vetter. "A 3D
        face model for pose and illumination invariant face recognition",
        IEEE International Conference on Advanced Video and Signal Based
        Surveillance, pp. 296-301, 2009.
    """

    def __init__(
        self, shape_model, texture_model, landmarks, holistic_features, diagonal
    ):
        self.shape_model = shape_model
        self.texture_model = texture_model
        self.holistic_features = holistic_features
        self.diagonal = diagonal
        # Find mapping that brings landmarks in correspondence with shape model
        (
            self.model_landmarks_index,
            self.landmarks,
        ) = find_correspondences_between_shapes(
            landmarks, self.shape_model.mean(), return_pointcloud=True
        )

    @property
    def n_vertices(self):
        """
        Returns the number of vertices of the shape model's trimesh.

        :type: `int`
        """
        return self.shape_model.template_instance.n_points

    @property
    def n_triangles(self):
        """
        Returns the number of triangles of the shape model's trimesh.

        :type: `int`
        """
        return self.shape_model.template_instance.n_tris

    def instance(
        self, shape_weights=None, texture_weights=None, landmark_group="landmarks"
    ):
        r"""
        Generates a novel Morphable Model instance given a set of shape and
        texture weights. If no weights are provided, then the mean Morphable
        Model instance is returned.

        Note that the texture generated is always clipped to the range(0-1).

        Parameters
        ----------
        shape_weights : ``(n_weights,)`` `ndarray` or `list` or ``None``, optional
            The weights of the shape model that will be used to create a novel
            shape instance. If ``None``, the weights are assumed to be zero,
            thus the mean shape is used.
        texture_weights : ``(n_weights,)`` `ndarray` or `list` or ``None``, optional
            The weights of the texture model that will be used to create a
            novel texture instance. If ``None``, the weights are assumed
            to be zero, thus the mean appearance is used.
        landmark_group : `str`, optional
            The group name that will be used for the sparse landmarks that
            will be attached to the returned instance.

        Returns
        -------
        instance : `menpo.shape.ColouredTriMesh`
            The coloured trimesh instance.
        """
        if shape_weights is None:
            shape_weights = np.zeros(self.shape_model.n_active_components)
        if texture_weights is None:
            texture_weights = np.zeros(self.texture_model.n_active_components)

        # Generate instance
        shape_instance = self.shape_model.instance(shape_weights)
        texture_instance = self.texture_model.instance(texture_weights)

        # Create and return trimesh
        return self._instance(shape_instance, texture_instance, landmark_group)

    def random_instance(self, landmark_group="__landmarks__"):
        r"""
        Generates a random instance of the Morphable Model.

        Parameters
        ----------
        landmark_group : `str`, optional
            The group name that will be used for the sparse landmarks that
            will be attached to the returned instance. Default is
            ``'__landmarks__'``.

        Returns
        -------
        instance : `menpo.shape.ColouredTriMesh`
            The coloured trimesh instance.
        """
        # TODO: this bit of logic should to be transferred down to PCAModel
        shape_weights = np.random.randn(self.shape_model.n_active_components)
        shape_instance = self.shape_model.instance(shape_weights)
        texture_weights = np.random.randn(self.texture_model.n_active_components)
        texture_instance = self.texture_model.instance(texture_weights)

        return self._instance(shape_instance, texture_instance, landmark_group)

    def __str__(self):
        cls_str = r"""{}
 - Shape model class: {}
   - {} vertices, {} triangles
   - {} shape components
   - Instance class: {}
 - Texture model class: {}
   - {} texture components
   - Diagonal of {} pixels
   - Features function is {}
   - {} channels
 - Sparse landmarks class: {}
   - {} landmarks
""".format(
            self._str_title,
            name_of_callable(self.shape_model),
            self.n_vertices,
            self.n_triangles,
            self.shape_model.n_components,
            name_of_callable(self.shape_model.template_instance),
            name_of_callable(self.texture_model),
            self.texture_model.n_components,
            self.diagonal,
            name_of_callable(self.holistic_features),
            self.n_channels,
            name_of_callable(self.landmarks),
            self.landmarks.n_points,
        )
        return cls_str


class ColouredMorphableModel(MorphableModel):
    r"""
    A Morphable Model where the texture information is assigned as a per-vertex
    Color vector.

    Parameters
    ----------
    shape_model : `menpo.model.PCAModel`
        The PCA model of the 3D shape. It is assumed that a shape instance is
        defined as a `menpo.shape.TriMesh`.
    texture_model : `menpo.model.PCAVectorModel`
        The PCA model of the per-vertex texture. It is assumed that a texture
        instance is defined as an ``(n_vertices * n_channels,)`` vector,
        where `n_vertices` should be the same as in the case of the
        `shape_model`. Note that the texture model can be feature based, in
        which case you need to specify the employed `holistic_features`.
    landmarks : `menpo.shape.PointUndirectedGraph`
        The set of sparse landmarks defined in the 3D space.
    holistic_features : `function`
        The features that were used for building the texture model. Please
        refer to `menpo.feature` for a list of potential features.
    diagonal : `int`
        This parameter was used to rescale the training images so that the
        diagonal of their bounding boxes matches the provided value. In other
        words, this parameter defined the size of the training images before
        extracting features.

    References
    ----------
    .. [1] V. Blanz, T. Vetter. "A morphable model for the synthesis of 3D
        faces", Conference on Computer Graphics and Interactive Techniques,
        pp. 187-194, 1999.
    .. [2] P. Paysan, R. Knothe, B. Amberg, S. Romdhani, T. Vetter. "A 3D
        face model for pose and illumination invariant face recognition",
        IEEE International Conference on Advanced Video and Signal Based
        Surveillance, pp. 296-301, 2009.
    """

    @property
    def _str_title(self):
        return "Coloured Morphable Model"

    @property
    def n_channels(self):
        """
        Returns the number of channels of the texture model.

        :type: `int`
        """
        return int(self.texture_model.n_features / self.n_vertices)

    def _instance(self, shape_instance, texture_instance, landmark_group):
        # Reshape the texture instance
        texture_instance = texture_instance.reshape([-1, self.n_channels])

        # restrict the texture to 0-1
        # texture_instance = np.clip(texture_instance, 0, 1)

        # Create trimesh
        trimesh = ColouredTriMesh(
            shape_instance.points,
            trilist=shape_instance.trilist,
            colours=texture_instance,
        )
        # Attach landmarks to trimesh
        trimesh.landmarks[landmark_group] = self.landmarks
        # Return trimesh
        return trimesh

    def sample_texture_model(self, bcoords, tri_indices):
        shape_template = self.shape_model.template_instance
        vertex_indices = shape_template.trilist[tri_indices]

        t_model = self.texture_model.components.reshape(
            [self.texture_model.n_active_components, -1, self.n_channels]
        )
        # n: components    s: samples    t: triangle    c: channels
        return np.einsum("nstc, st -> scn", t_model[:, vertex_indices], bcoords)

    def project_instance_on_texture_model(self, instance):
        return self.texture_model.project(instance.colours.ravel())


class TexturedMorphableModel(MorphableModel):
    r"""
    A Morphable Model where the texture information is assigned as image, with
    texture coordinates linking the vertex locations to the texture. This allows
    for a Model with differing spatial and texture resolutions.

    Parameters
    ----------
    shape_model : `menpo.model.PCAModel`
        The PCA model of the 3D shape. It is assumed that a shape instance is
        defined as a :map:`menpo.shape.TriMesh`.
    texture_model : :map:`PCAModel`.
        The PCA model of texture. It is assumed that a texture instance is
        defined as a `menpo.image.Image`. Note that the texture model can be
        feature based, in which case you need to specify the employed
        `holistic_features`.
    landmarks : `menpo.shape.PointUndirectedGraph`
        The set of sparse landmarks defined in the 3D space.
    tcoords : `menpo.shape.PointCloud`
        The texture coordinates linking any instance of the `texture_model`
        to any instance of the `shape_model`.
    holistic_features : `function`
        The features that were used for building the texture model. Please
        refer to `menpo.feature` for a list of potential features.
    diagonal : `int`
        This parameter was used to rescale the training images so that the
        diagonal of their bounding boxes matches the provided value. In other
        words, this parameter defined the size of the training images before
        extracting features.
    """

    def __init__(
        self,
        shape_model,
        texture_model,
        landmarks,
        tcoords,
        holistic_features,
        diagonal,
    ):
        super(TexturedMorphableModel, self).__init__(
            shape_model, texture_model, landmarks, holistic_features, diagonal
        )
        self.tcoords = tcoords
        self.tcoords_pixel_scaled = self.instance().tcoords_pixel_scaled()

    @property
    def _str_title(self):
        return "Textured Morphable Model"

    @property
    def n_channels(self):
        """
        Returns the number of channels of the texture model.

        :type: `int`
        """
        return self.texture_model.template_instance.n_channels

    def _instance(self, shape_instance, texture_instance, landmark_group):
        # Create trimesh

        # restrict the texture to 0-1
        # texture_instance.pixels = np.clip(texture_instance.pixels, 0, 1)

        trimesh = TexturedTriMesh(
            shape_instance.points,
            trilist=shape_instance.trilist,
            tcoords=self.tcoords.points,
            texture=texture_instance,
        )
        # Attach landmarks to trimesh
        trimesh.landmarks[landmark_group] = self.landmarks
        # Return trimesh
        return trimesh

    def sample_texture_model(self, bcoords, tri_indices):
        shape_template = self.shape_model.template_instance
        texture_template = self.texture_model.template_instance

        sample_points_in_texture = shape_template.barycentric_coordinate_interpolation(
            self.tcoords_pixel_scaled.points, bcoords, tri_indices
        )

        to_index = sample_points_in_texture.round().astype(int)

        texture_index_img = texture_template.from_vector(
            np.arange(self.texture_model.n_features)
        )

        # TODO shouldn't get out of mask samples here, but we do (hence
        # unmasked). This means for some samples every time in UV space
        # we are getting the wrong point.
        pca_sample_locations = texture_index_img.as_unmasked().sample(to_index, order=0)

        x = self.texture_model.components[:, pca_sample_locations]
        return np.swapaxes(x, 0, 2)

    def project_instance_on_texture_model(self, instance):
        return self.texture_model.project(instance.texture)


def find_correspondences_between_shapes(source, target, return_pointcloud=True):
    """
    It maps the landmarks in the source pointcloud to their corresponding
    indices in the target pointcloud. It optionally returns the source
    pointcloud with landmarks in correspondence to the target pointcloud.

    Parameters
    ----------
    source : `menpo.shape.PointCloud`
        The source pointcloud.
    target : `menpo.shape.PointCloud`
        The target pointcloud.
    return_pointcloud : `bool`, optional
        If ``True``, then the source pointcloud with landmarks in correspondence
        to the target pointcloud. It is returned only
        if ``return_pointcloud == True``.

    Returns
    -------
    map_source_to_target : ``(source_n_points,)`` `ndarray`
        The mapping between the points of source and target.
    source_in_correspondence : `menpo.shape.PointCloud`
        The source pointcloud with landmarks in correspondence will also be
        returned.
    """
    # Compute distance between source and target landmarks
    distances = source.distance_to(target)

    # Map source landmarks to target landmarks
    map_source_to_target = np.argmin(distances, 1)

    # Create pointcloud with source landmarks that are in correspondence
    # with target
    if return_pointcloud:
        return map_source_to_target, PointCloud(target.points[map_source_to_target, :])
    else:
        return map_source_to_target



import typing as t
from contextlib import contextmanager
from enum import IntEnum
from functools import cached_property, wraps
from pathlib import Path

import moderngl
import numpy as np
from menpo.image import MaskedImage
from menpo.shape import TriMesh
from menpo.transform import Homogeneous

from .transform import clip_to_image_transform

CONTAINING_DIR = Path(__file__).parent
PASSTHROUGH_TEXTURE_VERT_SHADER = CONTAINING_DIR / "shaders/texture/passthrough.vert"
PASSTHROUGH_TEXTURE_FRAG_SHADER = CONTAINING_DIR / "shaders/texture/passthrough.frag"
PASSTHROUGH_PER_VERTEX_VERT_SHADER = (
    CONTAINING_DIR / "shaders/per_vertex/passthrough.vert"
)
PASSTHROUGH_PER_VERTEX_FRAG_SHADER = (
    CONTAINING_DIR / "shaders/per_vertex/passthrough.frag"
)


class _LOADED_SHADER_TYPE(IntEnum):
    TEXTURE = 0
    PER_VERTEX = 1


def tri_bcoords_for_mesh(mesh):
    bc_per_tri = np.array([[1, 0], [0, 1], [0, 0]])
    bc = np.tile(bc_per_tri.T, mesh.n_tris).T

    index = np.repeat(np.arange(mesh.n_tris), 3, axis=0)

    return np.hstack((bc, index[:, None]))


def dedup_vertices(mesh):
    old_to_new = mesh.trilist.ravel()
    new_trilist = np.arange(old_to_new.shape[0]).reshape([-1, 3])
    new_points = mesh.points[old_to_new]
    return TriMesh(new_points, trilist=new_trilist), old_to_new


def _verify_opengl_homogeneous_matrix(matrix):
    if matrix.shape != (4, 4):
        raise ValueError("OpenGL matrices must have shape (4,4)")
    return np.require(matrix, dtype=np.float32, requirements="C")


@contextmanager
def safe_release(obj):
    try:
        yield obj
    except:
        obj.release()


def with_context(method):
    """
    Because it's possible to make multiple GLRasterizers in the same Python
    process it's important we ensure that the correct OpenGL context is activated
    when rasterization is performed. This is particularly important given that
    the rasterizers may be created inside Python processes that have OpenGL
    contexts not even owned by menpo3d (e.g. VTK or otherwise).
    """

    @wraps(method)
    def _impl(self, *method_args, **method_kwargs):
        with self.opengl_ctx:
            return method(self, *method_args, **method_kwargs)

    return _impl


class BasePassthroughProgram:
    def __init__(self, context):
        """
        Base program for a "passthrough" shader which just rasterizes the mesh
        without considering any lighting or mesh normals.

        Parameters
        ----------
        context : moderngl context
            The context used for rendering
        """
        self.context = context
        self.packed_vbo = None
        self.index_buffer_vbo = None

    @property
    def program(self):
        # Subclasses should lazily instantiate the program to ensure the GL
        # context is valid
        return None

    def __del__(self):
        # Make sure to always release the VBOs when the object is destroyed
        if self.packed_vbo is not None:
            self.packed_vbo.release()
        if self.index_buffer_vbo is not None:
            self.index_buffer_vbo.release()

    def _get_packed_vbo(self, buffer_bytes):
        # Handle creating the "packed" vbo which contains the data required
        # for rendering with the shader. buffer_bytes is a bytes array. This
        # tries to "cache" the VBO in the case that repeated rendering of the
        # same mesh is performed
        n_bytes = len(buffer_bytes)
        if self.packed_vbo is None or n_bytes != self.packed_vbo._size:
            if self.packed_vbo is not None:
                self.packed_vbo.release()
            self.packed_vbo = self.context.buffer(buffer_bytes)
        elif self.packed_vbo is not None:
            self.packed_vbo.write(buffer_bytes)

    def _get_index_buffer_vbo(self, buffer_bytes):
        # Handle creating the index buffer vbo which contains the data required
        # for rendering with the shader. buffer_bytes is a bytes array. This
        # tries to "cache" the VBO in the case that repeated rendering of the
        # same mesh is performed
        n_bytes = len(buffer_bytes)
        if self.index_buffer_vbo is None or n_bytes != self.index_buffer_vbo._size:
            if self.index_buffer_vbo is not None:
                self.index_buffer_vbo.release()
            self.index_buffer_vbo = self.context.buffer(buffer_bytes)
        elif self.index_buffer_vbo is not None:
            self.index_buffer_vbo.write(buffer_bytes)

    def set_mvp(self, mvp_matrix):
        """
        Set the MVP (Model, View, Projection) matrix for the shader. This is
        modelled as a uniform called "MVP".

        Parameters
        ----------
        mvp_matrix : ndarray (4, 4)
            The combined (Model, View, Projection) matrix to set
        """
        mvp_matrix = np.require(mvp_matrix, dtype=np.float32, requirements=["C"])
        self.program["MVP"].write(mvp_matrix.tobytes())


class TexturePassthroughProgram(BasePassthroughProgram):
    def __init__(self, context):
        """
        Implements a "passthrough" shader for textured meshes. This just
        rasterizes the mesh without considering any lighting or mesh normals.

        Parameters
        ----------
        context : moderngl context
            The context used for rendering
        """
        super().__init__(context)
        self.texture = None
        self._texture_shape = None
        self._pixels_ref = None

    @cached_property
    def program(self):
        return self.context.program(
            vertex_shader=PASSTHROUGH_TEXTURE_VERT_SHADER.read_text(encoding="ascii"),
            fragment_shader=PASSTHROUGH_TEXTURE_FRAG_SHADER.read_text(encoding="ascii"),
        )

    def __del__(self):
        super().__del__()
        if self.texture is not None:
            self.texture.release()
        # Release program last
        if self.program is not None:
            self.program.release()

    def _build_texture(self, pixels):
        assert pixels.shape[-1] == 3, "Only RGB textures supported"

        # Try caching the texture if repeated rendering of the same mesh is
        # being performed
        shape = pixels.shape[:2][::-1]
        if self._texture_shape is None or self._texture_shape != shape:
            self._texture_shape = shape
            # Note this keeps the texture alive and so uses lots of memory for large textures
            self._pixels_ref = pixels
            self.texture = self.context.texture(
                size=shape, components=3, data=pixels.tobytes(), dtype="f4"
            )
        elif not np.allclose(self._pixels_ref, pixels):
            assert self.texture is not None
            self.texture.write(pixels.tobytes())

    def create_vao(self, mesh, per_vertex_f3v):
        """
        Create a simple Vertex Array Object (VAO) that contains the per vertex
        array data used by the shader. Ensures that the correct texture
        is also allocated.

        Parameters
        ----------
        mesh :  `menpo.shape.TexturedTriMesh`
        per_vertex_f3v : ndarray, (n_points, 3)

        Returns
        -------
        `moderngl.vertex_array.VertexArray`
            VAO to be used by the moderngl rendering context
        """
        # Grab the texture in [H, W, 3] format - note the pixel values need to
        # be [0, 1] for moderngl
        pixels = mesh.texture.pixels_with_channels_at_back(out_dtype=np.float32)
        # Note that we assume the texture coordinates are flipped inside the
        # shader
        self._build_texture(pixels)
        # Create the packed VBO where the entries here are in the same order as
        # the shader inputs
        packed = np.concatenate(
            [mesh.points, mesh.tcoords.points, per_vertex_f3v], axis=1
        )

        packed_buffer = np.require(packed, dtype=np.float32, requirements=["C"])
        index_buffer = np.require(mesh.trilist, dtype=np.uint32, requirements=["C"])
        self._get_packed_vbo(packed_buffer.tobytes())
        self._get_index_buffer_vbo(index_buffer.tobytes())
        # Note the ordering of the inputs list matches the ordering of the
        # packed vector above
        return self.context.simple_vertex_array(
            self.program,
            self.packed_vbo,
            "in_vert",
            "in_text",
            "in_f3v",
            index_buffer=self.index_buffer_vbo,
        )


class PerVertexPassthroughProgram(BasePassthroughProgram):
    def __init__(self, context):
        """
        Implements a "passthrough" shader for coloured meshes. This just
        rasterizes the mesh without considering any lighting or mesh normals.

        Parameters
        ----------
        context : moderngl context
            The context used for rendering
        """
        super().__init__(context)

    @cached_property
    def program(self):
        return self.context.program(
            vertex_shader=PASSTHROUGH_PER_VERTEX_VERT_SHADER.read_text(
                encoding="ascii"
            ),
            fragment_shader=PASSTHROUGH_PER_VERTEX_FRAG_SHADER.read_text(
                encoding="ascii"
            ),
        )

    def __del__(self):
        super().__del__()
        # Release program last
        if self.program is not None:
            self.program.release()

    def create_vao(self, mesh, per_vertex_f3v):
        """
        Create a simple Vertex Array Object (VAO) that contains the per vertex
        array data used by the shader. Ensures that the per-vertex colour has
        the correct format (n_points, 3). Note that if no per-vertex colour
        is passed then a uniform gray colour is used (0.5, 0.5, 0.5).

        Parameters
        ----------
        mesh :  `menpo.shape.ColouredTriMesh`
        per_vertex_f3v : ndarray, (n_points, 3)

        Returns
        -------
        `moderngl.vertex_array.VertexArray`
            VAO to be used by the moderngl rendering context
        """
        colours = mesh.colours
        if colours is None:
            # Default to gray per vertex colouring
            colours = np.full(mesh.points.shape, 0.5, dtype=np.float32)
        if colours.shape[-1] == 1:
            # Force from grayscale to colour by repeating
            colours = np.repeat(colours, 3, axis=-1)
        assert colours.shape[-1] == 3, "Only RGB colours are supported"

        packed = np.concatenate([mesh.points, colours, per_vertex_f3v], axis=1)

        packed_buffer = np.require(packed, dtype=np.float32, requirements=["C"])
        index_buffer = np.require(mesh.trilist, dtype=np.uint32, requirements=["C"])
        self._get_packed_vbo(packed_buffer.tobytes())
        self._get_index_buffer_vbo(index_buffer.tobytes())
        return self.context.simple_vertex_array(
            self.program,
            self.packed_vbo,
            "in_vert",
            "in_color",
            "in_f3v",
            index_buffer=self.index_buffer_vbo,
        )


class GLRasterizer:
    def __init__(
        self,
        width: int = 1024,
        height: int = 768,
        model_matrix: t.Optional[np.ndarray] = None,
        view_matrix: t.Optional[np.ndarray] = None,
        projection_matrix: t.Optional[np.ndarray] = None,
        opengl_context: t.Optional[moderngl.Context] = None,
    ) -> None:
        # Make a single OpenGL context that will be managed by the lifetime of
        # this class. We will dynamically create two default "pass through"
        # programs based on the type of the input mesh
        self.opengl_ctx = opengl_context or moderngl.create_standalone_context()
        self.width = width
        self.height = height

        self._model_matrix = model_matrix if model_matrix is not None else np.eye(4)
        self._view_matrix = view_matrix if view_matrix is not None else np.eye(4)
        self._projection_matrix = (
            projection_matrix if projection_matrix is not None else np.eye(4)
        )
        self._vertex_shader = None
        self._fragment_shader = None
        self._shader_type = None
        # We will dynamically build the program based on the mesh type
        self._active_program = None
        self._texture_program = None
        self._per_vertex_program = None

    @cached_property
    def _f3v_renderbuffer(self):
        # Lazily instantiate all buffers to ensure the GL context is active
        return self.opengl_ctx.renderbuffer(self.size, components=3, dtype="f4")

    @cached_property
    def _rgba_renderbuffer(self):
        # Lazily instantiate all buffers to ensure the GL context is active
        return self.opengl_ctx.renderbuffer(self.size, components=4, dtype="f4")

    @cached_property
    def _depth_renderbuffer(self):
        # Lazily instantiate all buffers to ensure the GL context is active
        return self.opengl_ctx.depth_renderbuffer(self.size)

    @cached_property
    def _fbo(self):
        # Lazily instantiate all buffers to ensure the GL context is active
        return self.opengl_ctx.framebuffer(
            [self._rgba_renderbuffer, self._f3v_renderbuffer],
            self._depth_renderbuffer,
        )

    def __del__(self):
        # Ensure we avoid memory leaks by manually releasing all moderngl
        # objects
        self._rgba_renderbuffer.release()
        self._f3v_renderbuffer.release()
        self._depth_renderbuffer.release()

    @property
    def size(self):
        return self.width, self.height

    @property
    def model_matrix(self):
        return self._model_matrix

    @model_matrix.setter
    def model_matrix(self, value):
        self._model_matrix = _verify_opengl_homogeneous_matrix(value)

    @property
    def view_matrix(self):
        return self._view_matrix

    @view_matrix.setter
    def view_matrix(self, value):
        self._view_matrix = _verify_opengl_homogeneous_matrix(value)

    @property
    def projection_matrix(self):
        return self._projection_matrix

    @projection_matrix.setter
    def projection_matrix(self, value):
        self._projection_matrix = _verify_opengl_homogeneous_matrix(value)

    @property
    def mvp_matrix(self):
        return np.linalg.multi_dot(
            [self.projection_matrix, self.view_matrix, self.model_matrix]
        )

    @property
    def model_to_clip_matrix(self):
        return np.dot(
            self.projection_matrix, np.dot(self.view_matrix, self.model_matrix)
        )

    @property
    def model_transform(self):
        return Homogeneous(self.model_matrix)

    @property
    def view_transform(self):
        return Homogeneous(self.view_matrix)

    @property
    def projection_transform(self):
        return Homogeneous(self.projection_matrix)

    @property
    def mvp_transform(self):
        return Homogeneous(self.mvp_matrix)

    @property
    def model_to_clip_transform(self):
        r"""
        Transform that takes 3D points from model space to 3D clip space
        """
        return Homogeneous(self.model_to_clip_matrix)

    @property
    def clip_to_image_transform(self):
        r"""
        Affine transform that converts 3D clip space coordinates into 2D image
        space coordinates
        """
        return clip_to_image_transform(self.width, self.height)

    @property
    def model_to_image_transform(self):
        r"""
        TransformChain from 3D model space to 2D image space.
        """
        return self.model_to_clip_transform.compose_before(self.clip_to_image_transform)

    def _set_active_program_to_texture(self):
        # Ensure that we only build a single texture program per rasterizer
        if self._texture_program is None:
            self._texture_program = TexturePassthroughProgram(self.opengl_ctx)
        self._active_program = self._texture_program
        self._shader_type = _LOADED_SHADER_TYPE.TEXTURE

    def _set_active_program_to_per_vertex(self):
        # Ensure that we only build a single per vertex program per rasterizer
        if self._per_vertex_program is None:
            self._per_vertex_program = PerVertexPassthroughProgram(self.opengl_ctx)
        self._active_program = self._per_vertex_program
        self._shader_type = _LOADED_SHADER_TYPE.PER_VERTEX

    def _set_active_program_by_mesh_type(self, mesh):
        # Set the active program based on the input mesh type - this allows
        # lazy instantiation of the program
        if hasattr(mesh, "tcoords"):
            self._set_active_program_to_texture()
        elif hasattr(mesh, "colours"):
            self._set_active_program_to_per_vertex()
        else:
            raise ValueError(
                "Unknown mesh type, only textured (tcoords) or "
                "coloured (colours) TriMeshes are supported"
            )

    @with_context
    def _rasterize(self, mesh, per_vertex_f3v, fetch_f3v=True):
        """
        This defines the main rasterize method with moderngl. Ensures that
        the program is setup correctly and that the outputs are also parsed
        correctly.

        Parameters
        ----------
        mesh : `menpo.shape.ColouredTriMesh` or `menpo.shape.TexturedTriMesh`
            Mesh to render
        per_vertex_f3v : ndarray, (n_points, 3)
            Per vertex floats to render
        fetch_f3v : bool
            If True, then fetch the per-vertex floats, otherwise skip fetching them.
            Note that at the moment they are always rendered but we can skip
            fetching them if we don't use them.

        Returns
        -------
        rgb_image : ndarray float32, [H, W, 3]
            RGB image representing the rasterized image (with either the
            texture rasterized or the per vertex colours)
        f3v_image : ndarray float32, [H, W, 3] or None if fetch_f3v is False
            RGB image representing the rasterized per-vertex arbitrary floating
            point values
        mask : ndarray bool, [H, W, 1]
            True where pixels were written to, False otherwise
        """
        self._active_program.set_mvp(self.mvp_matrix)
        vao = self._active_program.create_vao(mesh, per_vertex_f3v)

        # Rendering
        self._fbo.use()
        self.opengl_ctx.clear(1.0, 1.0, 1.0, 0.0)
        self.opengl_ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        if self._shader_type == _LOADED_SHADER_TYPE.TEXTURE:
            self._active_program.texture.use()

        with safe_release(vao):
            vao.render()

        # Snure that the context has finished drawing (block) before we start
        # to read from buffers
        self.opengl_ctx.finish()

        # Note that this is dependent on the shader output locations and
        # how the renderbuffers are wired up
        rgba_data = self._fbo.read(components=4, attachment=0, dtype="f4")
        f3v_data = None
        if fetch_f3v:
            f3v_data = self._fbo.read(components=3, attachment=1, dtype="f4")

        rgba_image = np.frombuffer(rgba_data, dtype=np.float32)
        rgba_image = rgba_image.reshape(self.height, self.width, 4)[::-1]
        rgba_image = np.require(rgba_image, dtype=np.float32, requirements=["C"])

        f3v_image = None
        if f3v_data is not None:
            f3v_image = np.frombuffer(f3v_data, dtype=np.float32)
            f3v_image = f3v_image.reshape(self.height, self.width, 3)[::-1]
            f3v_image = np.require(f3v_image, dtype=np.float32, requirements=["C"])

        return rgba_image[..., :3], f3v_image, rgba_image[..., -1].astype(np.bool)

    def rasterize_mesh_with_f3v_interpolant(self, mesh, per_vertex_f3v=None):
        r"""
        Rasterize the object to an image and generate an interpolated
        3-float image from a per vertex float 3 vector.

        Parameters
        ----------
        mesh : object implementing the Rasterizable interface.
        per_vertex_f3v : optional, ndarray (n_points, 3)
            A per-vertex 3 vector of floats that will be interpolated across
            the image.
            If None, a zero is passed for every vertex.

        Returns
        -------
        rgb_image : 3-channel MaskedImage of shape (width, height)
            The result of the rasterization. Mask is true iff the pixel was
            rendered to by OpenGL.

        interp_image: 3 channel MaskedImage of shape (width, height)
            The result of interpolating the per_vertex_f3v across the
            visible primitives.

        """
        if not (hasattr(mesh, "points") and hasattr(mesh, "trilist")):
            raise ValueError(
                "Rasterizable types must have points and trilist properties."
            )

        if per_vertex_f3v is None:
            per_vertex_f3v = np.zeros(mesh.points.shape, dtype=np.float32)

        self._set_active_program_by_mesh_type(mesh)

        rgb_pixels, f3v_pixels, mask = self._rasterize(mesh, per_vertex_f3v)
        images = (
            MaskedImage.init_from_channels_at_back(rgb_pixels, mask=mask),
            MaskedImage.init_from_channels_at_back(f3v_pixels, mask=mask),
        )

        # Transform all landmarks and set them on the image
        image_lms = self.model_to_image_transform.apply(mesh.landmarks)
        for image in images:
            image.landmarks = image_lms

        return images

    def rasterize_mesh_with_shape_image(self, mesh):
        r"""Rasterize a mesh and additionally generate an interpolated
        3-float image from the shape information on the mesh.

        Parameters
        ----------
        mesh : object implementing the Rasterizable interface.

        Returns
        -------
        rgb_image : 3 channel MaskedImage of shape (width, height)
            The result of the rasterization. Mask is true iff the pixel was
            rendered to by OpenGL.
        shape_image: 3 channel MaskedImage of shape (width, height)
            The result of interpolating the spatial information of each vertex
            across the visible primitives. Note that the shape information
            is *NOT* adjusted by the P,V,M matrices, and so the resulting
            shape image is always in the original objects reference shape
            (i.e. the z value will not necessarily correspond to a depth
            buffer).
        """
        return self.rasterize_mesh_with_f3v_interpolant(
            mesh, per_vertex_f3v=mesh.points
        )

    def rasterize_mesh(self, mesh):
        r"""Rasterize a mesh to an image.

        Parameters
        ----------
        mesh : object implementing the Rasterizable interface.

        Returns
        -------
        rgb_image : 3 channel MaskedImage of shape (width, height)
            The result of the rasterization. Mask is true iff the pixel was
            rendered to by OpenGL.
        """
        self._set_active_program_by_mesh_type(mesh)

        per_vertex_f3v = np.zeros(mesh.points.shape, dtype=np.float32)
        rgb_pixels, _, mask = self._rasterize(mesh, per_vertex_f3v, fetch_f3v=False)
        image = MaskedImage.init_from_channels_at_back(rgb_pixels, mask=mask)
        image.landmarks = self.model_to_image_transform.apply(mesh.landmarks)

        return image

    def rasterize_barycentric_coordinate_image(self, mesh):
        # Convert the mesh into a version with one vertex per triangle
        mesh, _ = dedup_vertices(mesh)

        per_vertex_f3v = tri_bcoords_for_mesh(mesh)

        images = self.rasterize_mesh_with_f3v_interpolant(
            mesh, per_vertex_f3v=per_vertex_f3v
        )

        # the interpolated image is [tri_index, alpha, beta]
        # -> split this into two images, one tri_index, one bc
        inverse_image = images[1]

        vectors = inverse_image.as_vector(keep_channels=True)
        tri_indices = vectors[2].astype(np.uint32)

        a, b = vectors[:2]
        g = 1 - a - b
        bcoords = np.vstack([a, b, g])

        tri_index_image = inverse_image.from_vector(tri_indices, n_channels=1)
        bcoords_image = inverse_image.from_vector(bcoords, n_channels=3)

        return tri_index_image, bcoords_image



import menpo
import menpo3d.io as mio
import numpy as np
from menpo.image import Image
from menpo.shape import ColouredTriMesh, PointCloud, TexturedTriMesh, TriMesh
from numpy.testing import assert_allclose

# ground truth bunny landmarks
from pytest import raises

bunny_leye = np.array(
    [[-0.0907334573249135, 0.13944519477304135, 0.016432549244098166]]
)
bunny_reye = np.array([[-0.06188841143305087, 0.1404748336910087, 0.03628544974441027]])
bunny_nose = np.array(
    [[-0.0824712814601403, 0.12724167358964233, 0.051332619501298894]]
)
bunny_mouth = np.array(
    [
        [-0.08719271323528464, 0.11440556680892036, 0.044485263772198975],
        [-0.0733831000935048, 0.11610087742529278, 0.051836298703986136],
    ]
)


def test_import_asset_bunny():
    mesh = mio.import_builtin_asset("bunny.obj")
    assert isinstance(mesh, TriMesh)
    assert isinstance(mesh.points, np.ndarray)
    assert mesh.points.shape[1] == 3
    assert isinstance(mesh.trilist, np.ndarray)
    assert mesh.trilist.shape[1] == 3


def test_import_asset_james():
    mesh = mio.import_builtin_asset("james.obj")
    assert isinstance(mesh, TexturedTriMesh)
    assert isinstance(mesh.points, np.ndarray)
    assert mesh.points.shape[1] == 3
    assert isinstance(mesh.trilist, np.ndarray)
    assert mesh.trilist.shape[1] == 3
    assert isinstance(mesh.texture, Image)
    assert isinstance(mesh.tcoords, PointCloud)
    assert mesh.tcoords.points.shape[1] == 2


def test_import_asset_template():
    mesh = mio.import_builtin_asset("template.ply")
    assert isinstance(mesh, ColouredTriMesh)
    assert isinstance(mesh.points, np.ndarray)
    assert mesh.points.shape[1] == 3
    assert isinstance(mesh.trilist, np.ndarray)
    assert mesh.trilist.shape[1] == 3
    assert isinstance(mesh.colours, np.ndarray)
    assert mesh.colours.shape[1] == 3
    assert mesh.colours.max() == 1.0


def test_import_incorrect_built_in():
    with raises(ValueError):
        mio.import_builtin_asset("adskljasdlkajd.obj")


def test_json_landmarks_bunny():
    mesh = mio.import_builtin_asset("bunny.obj")
    assert "LJSON" in mesh.landmarks.group_labels
    lms = mesh.landmarks["LJSON"]
    labels = {"reye", "mouth", "nose", "leye"}
    assert len(labels - set(lms.labels)) == 0
    assert_allclose(lms.with_labels("leye").points, bunny_leye, atol=1e-7)
    assert_allclose(lms.with_labels("reye").points, bunny_reye, atol=1e-7)
    assert_allclose(lms.with_labels("nose").points, bunny_nose, atol=1e-7)
    assert_allclose(lms.with_labels("mouth").points, bunny_mouth, atol=1e-7)


def test_custom_landmark_logic_bunny():
    def f(path):
        lmark_dict = {
            "no_nose": path.with_name("bunny_no_nose.ljson"),
            "full_set": path.with_name("bunny.ljson"),
        }
        return menpo.io.input.resolve_from_paths(lmark_dict)

    mesh = mio.import_mesh(mio.data_path_to("bunny.obj"), landmark_resolver=f)
    assert "no_nose" in mesh.landmarks.group_labels
    lms = mesh.landmarks["no_nose"]
    labels = {"reye", "mouth", "leye"}
    assert len(set(lms.labels) - labels) == 0
    assert_allclose(lms.with_labels("leye").points, bunny_leye, atol=1e-7)
    assert_allclose(lms.with_labels("reye").points, bunny_reye, atol=1e-7)
    assert_allclose(lms.with_labels("mouth").points, bunny_mouth, atol=1e-7)

    assert "full_set" in mesh.landmarks.group_labels
    lms = mesh.landmarks["full_set"]
    labels = {"reye", "mouth", "nose", "leye"}
    assert len(set(lms.labels) - labels) == 0
    assert_allclose(lms.with_labels("leye").points, bunny_leye, atol=1e-7)
    assert_allclose(lms.with_labels("reye").points, bunny_reye, atol=1e-7)
    assert_allclose(lms.with_labels("nose").points, bunny_nose, atol=1e-7)
    assert_allclose(lms.with_labels("mouth").points, bunny_mouth, atol=1e-7)


def test_custom_landmark_logic_None_bunny():
    def f(mesh):
        return None

    mesh = mio.import_mesh(mio.data_path_to("bunny.obj"), landmark_resolver=f)
    assert mesh.landmarks.n_groups == 0


def test_json_landmarks_bunny_direct():
    lms = mio.import_landmark_file(mio.data_path_to("bunny.ljson"))["LJSON"]
    labels = {"reye", "mouth", "nose", "leye"}
    assert len(labels - set(lms.labels)) == 0
    assert_allclose(lms.with_labels("leye").points, bunny_leye, atol=1e-7)
    assert_allclose(lms.with_labels("reye").points, bunny_reye, atol=1e-7)
    assert_allclose(lms.with_labels("nose").points, bunny_nose, atol=1e-7)
    assert_allclose(lms.with_labels("mouth").points, bunny_mouth, atol=1e-7)


def test_ls_builtin_assets():
    assert set(mio.ls_builtin_assets()) == {
        "bunny.ljson",
        "bunny.obj",
        "bunny_no_nose.ljson",
        "james.jpg",
        "james.ljson",
        "james.obj.mtl",
        "james.obj",
        "template.ljson",
        "template.ply",
    }


def test_mesh_paths():
    ls = mio.mesh_paths(mio.data_dir_path())
    assert len(list(ls)) == 3


def test_import_landmark_files_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_landmark_files("asldfjalkgjlaknglkajlekjaltknlaekstjlakj"))


def test_import_meshes_wrong_path_raises_value_error():
    with raises(ValueError):
        list(mio.import_meshes("asldfjalkgjlaknglkajlekjaltknlaekstjlakj"))



import contextlib
import os
import tempfile
from unittest.mock import MagicMock, PropertyMock, patch

import menpo3d.io as mio

test_obj = mio.import_builtin_asset("james.obj")
test_lg = mio.import_landmark_file(mio.data_path_to("bunny.ljson"))


@contextlib.contextmanager
def _temporary_path(extension):
    # Create a temporary file and remove it
    fake_path = tempfile.NamedTemporaryFile(delete=False, suffix=extension)
    fake_path.close()
    fake_path = fake_path.name
    os.unlink(fake_path)
    yield fake_path
    if os.path.exists(fake_path):
        os.unlink(fake_path)


@patch("menpo3d.io.output.base.Path.exists")
@patch("{}.open".format(__name__), create=True)
def test_export_mesh_obj(mock_open, exists):
    exists.return_value = False
    fake_path = "/fake/fake.obj"
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_mesh(test_obj, f, extension="obj")


def test_export_mesh_ply_ascii():
    with _temporary_path(".ply") as f:
        mio.export_mesh(test_obj, f)
        assert os.path.exists(f)


def test_export_mesh_ply_binary():
    with _temporary_path(".ply") as f:
        mio.export_mesh(test_obj, f, binary=True)
        assert os.path.exists(f)


@patch("PIL.Image.EXTENSION")
@patch("menpo.image.base.PILImage")
@patch("menpo3d.io.output.base.Path.exists")
@patch("menpo.io.output.base.Path.open")
def test_export_mesh_obj_textured(mock_open, exists, PILImage, PIL):
    PIL.return_value.Image.EXTENSION = {".jpg": None}
    exists.return_value = False
    mock_open.return_value = MagicMock()
    fake_path = "/fake/fake.obj"
    mio.export_textured_mesh(test_obj, fake_path, extension="obj")
    assert PILImage.fromarray.called


@patch("PIL.Image.EXTENSION")
@patch("menpo.image.base.PILImage")
@patch("menpo3d.io.output.base.Path.exists")
@patch("menpo.io.output.base.Path.open")
def test_export_mesh_ply_textured(mock_open, exists, PILImage, PIL):
    PIL.return_value.Image.EXTENSION = {".jpg": None}
    exists.return_value = False
    mock_open.return_value = MagicMock()
    with _temporary_path(".ply") as f:
        mio.export_textured_mesh(test_obj, f)
        assert PILImage.fromarray.called


@patch("menpo.io.output.landmark.json.dump")
@patch("menpo3d.io.output.base.Path.exists")
@patch("{}.open".format(__name__), create=True)
def test_export_landmark_ljson(mock_open, exists, json_dump):
    exists.return_value = False
    fake_path = "/fake/fake.ljson"
    with open(fake_path) as f:
        type(f).name = PropertyMock(return_value=fake_path)
        mio.export_landmark_file(test_lg, f, extension="ljson")
    assert json_dump.called



import numpy as np
from menpo.transform import Transform, Translation
from .math import radial_fit


class CylindricalUnwrap(Transform):
    r"""
    Maps 3D points :math:`(x, y, z)` into cylindrical coordinates
    :math:`(x', y', z')`

    .. math::

        x' &\leftarrow r \theta \\
        y' &\leftarrow y \\
        z' &\leftarrow d

    where

    .. math::

        d &= \sqrt{x^2 + y^2} - r \\
        \theta &= \arctan{\left(\frac{x}{z}\right)}

    The cylinder is oriented such that it's axial vector is ``[0, 1, 0]``
    and it's centre is at the origin. Discontinuity in :math:`\theta` values
    occurs at ``y-z`` plane for *negative* ``z`` values (i.e. the interesting
    information you are wanting to preserve in the unwrapping better have
    positive ``z`` values).

    Parameters
    ----------

    radius : `float`
        The distance of the unwrapping from the axis.

    """

    def __init__(self, radius):
        self.radius = radius

    def _apply(self, x, **kwargs):
        cy_coords = np.zeros_like(x)
        depth = np.sqrt(x[:, 0] ** 2 + x[:, 2] ** 2) - self.radius
        theta = np.arctan2(x[:, 0], x[:, 2])
        z = x[:, 1]
        cy_coords[:, 0] = theta * self.radius
        cy_coords[:, 1] = z
        cy_coords[:, 2] = depth
        return cy_coords


def optimal_cylindrical_unwrap(points):
    r"""
    Returns a :map:`TransformChain` of
    [:map:`Translation`, :map:`CylindricalUnwrap`]
    which optimally cylindrically unwraps the points provided. This is done by:

    #. Find an optimal :map:`Translation` to centre the points in ``x-z`` plane
    #. Use :map:`circle_fit` to find the optimal radius for fitting the points
    #. Calculate a :map:`CylindricalUnwrap` using the optimal radius
    #. Return a composition of the two.

    Parameters
    ----------
    points : :map:`PointCloud`
        The 3D points that will be used to find the optimum unwrapping position

    Returns
    -------

    transform: :map:`TransformChain`
        A :map:`TransformChain` which performs the optimal translation and
        unwrapping.

    """
    # find the optimum centre to unwrap
    xy = points.points[:, [0, 2]]  # just in the x-z plane
    centre, radius = radial_fit(xy)
    # convert the 2D circle data into the 3D space
    translation = np.array([centre[0], 0, centre[1]])
    centring_transform = Translation(-translation)
    unwrap = CylindricalUnwrap(radius)
    return centring_transform.compose_before(unwrap)



import numpy as np

# TODO: Should inherit from PCAModel (?)
class MultilinearModel(object):
    """Usage for a morphable model:

    shape_model = MultilinearModel(identity_model, expression_model)
    mm = TexturedMorphableModel(shape_model, texture_model, landmarks, tcoords, no_op, None)
    """

    def __init__(self, *models):
        self.models = models

    @property
    def n_active_components(self):
        return np.sum([mdl.n_active_components for mdl in self.models])

    @n_active_components.setter
    def n_active_components(self, values):
        try:
            for mdl, v in zip(self.models, values):
                mdl.n_active_components = value
        except TypeError:
            self.models[0].n_active_components = values

    def noise_variance(self):
        return self.models[0].noise_variance()

    def mean(self):
        v = np.sum([mdl.mean_vector for mdl in self.models], 0)

        return self.template_instance.from_vector(v)

    def instance(self, weights, normalized_weights=False):
        n_active = np.cumsum([0] + [mdl.n_active_components for mdl in self.models])
        slices = (slice(*s) for s in zip(n_active, n_active[1:]))

        v = np.sum(
            [
                mdl.instance_vector(weights[start_end], normalized_weights)
                for mdl, start_end in zip(self.models, slices)
            ],
            0,
        )

        return self.template_instance.from_vector(v)

    @property
    def components(self):
        return np.concatenate([mdl.components for mdl in self.models], 0)

    @property
    def template_instance(self):
        # TODO: Assert all models are of the same template instance.
        return self.models[0].template_instance

    @property
    def eigenvalues(self):
        return np.concatenate([mdl.eigenvalues for mdl in self.models], 0)

    def project(self, v):
        # Warning: Hacky way to project. Only projects onto the first
        # models components and sets the rest to zero.

        n_active_components = [mdl.n_active_components for mdl in self.models]
        e = np.zeros(sum(n_active_components[1:]))
        return np.hstack([self.models[0].project(v), e])



import menpo3d
import numpy as np
from menpo.shape import TriMesh
from menpo3d.vtkutils import trimesh_to_vtk, trimesh_from_vtk
from numpy.testing import assert_allclose
from pytest import raises


def test_trimesh_to_vtk_and_back_is_same():
    bunny = menpo3d.io.import_builtin_asset.bunny_obj()
    bunny_vtk = trimesh_to_vtk(bunny)
    bunny_back = trimesh_from_vtk(bunny_vtk)
    assert_allclose(bunny.points, bunny_back.points)
    assert np.all(bunny.trilist == bunny_back.trilist)


def test_trimesh_to_vtk_fails_on_2d_mesh():
    with raises(ValueError):
        points = np.random.random((5, 2))
        test_mesh = TriMesh(points)
        trimesh_to_vtk(test_mesh)


def test_barycentric_rebuild_returns_same_as_snapped_points():
    mesh = menpo3d.io.import_builtin_asset.bunny_obj()
    lms = mesh.landmarks[None]
    bc = mesh.barycentric_coordinates_of_pointcloud(lms)
    recon_lms = mesh.project_barycentric_coordinates(*bc)
    direct_recon_lms = mesh.snap_pointcloud_to_surface(lms)[0]
    assert_allclose(recon_lms.points, direct_recon_lms.points)



import numpy as np
from menpo.transform import Homogeneous, Rotation, Translation
from menpo3d.camera import pinhole_intrinsic_matrix
from menpo3d.rasterize.transform import flip_xy_yx

# Note this is copied from OpenCV to avoid importing OpenCV at the module level
SOLVEPNP_ITERATIVE = 0

_drop_h = Homogeneous(np.eye(4)[:3])


def solve_pnp(
    points_2d,
    points_3d,
    intrinsic_matrix,
    distortion_coefficients=None,
    pnp_method=SOLVEPNP_ITERATIVE,
    n_iterations=100,
    reprojection_error=8.0,
    initial_transform=None,
):
    """
    Use OpenCV to solve the Perspective-N-Point problem (PnP). Uses Ransac PnP as this
    typically provides better results. The image and mesh must both have the same
    landmark group name attached.

    Note the intrinsic matrix (if given) must be in "OpenCV" space and thus
    has the "x" and "y" axes flipped w.r.t the menpo norm. E.g. the intrinsic matrix
    is defined as follows:

        [fx,  0, cx, 0]
        [ 0, fy, cy, 0]
        [ 0,  0,  1, 0]
        [ 0,  0,  0, 1]

    Parameters
    ----------
    points_2d : :map`Pointcloud` or subclass
        The 2D points in the image to solve the PnP problem with.
    points_3d : :map`Pointcloud` or subclass
        The 3D points to solve the PnP problem with
    group : str, optional
        The name of the landmark group
    intrinsic_matrix : :map`Homogeneous`
        The intrinsic matrix - if the intrinsic matrix is unknow please see
        usage of pinhole_intrinsic_matrix()
    distortion_coefficients : ``(D,)`` `ndarray`
        The distortion coefficients (if not given assumes 0 coefficients). See the
        OpenCV documentation for the distortion coefficient types that are supported.
    pnp_method : int
        The OpenCV PNP method e.g. cv2.SOLVEPNP_ITERATIVE or otherwise
    n_iterations : int
        The number of iterations to perform
    reprojection_error : float
        The maximum reprojection error to allow for a point to be considered an
        inlier.
    initial_transform : :map`Homogeneous`
        The initialization for the cv2.SOLVEPNP_ITERATIVE method. Compatible
        with the returned model transformation returned by this method.

    Returns
    -------
    model_view_t : :map`Homogeneous`
        The combined ModelView transform. Can be used to place the 3D points
        in "eye space".
    proj_t : :map`Homogeneous`
        A transform that can be used to project the input 3D points
        back into the image
    """
    import cv2

    if distortion_coefficients is None:
        distortion_coefficients = np.zeros(4)

    r_vec = t_vec = None
    if initial_transform is not None:
        if pnp_method != cv2.SOLVEPNP_ITERATIVE:
            raise ValueError(
                "Initial estimates can only be given to SOLVEPNP_ITERATIVE"
            )
        else:
            r_vec = cv2.Rodrigues(initial_transform.h_matrix[:3, :3])[0]
            t_vec = initial_transform.h_matrix[:3, -1].ravel()

    converged, r_vec, t_vec, _ = cv2.solvePnPRansac(
        points_3d.points,
        points_2d.points[:, ::-1],
        intrinsic_matrix.h_matrix[:3, :3],
        distortion_coefficients,
        flags=pnp_method,
        iterationsCount=n_iterations,
        reprojectionError=reprojection_error,
        useExtrinsicGuess=r_vec is not None,
        rvec=r_vec,
        tvec=t_vec,
    )
    if not converged:
        raise ValueError("cv2.solvePnPRansac failed to converge")

    rotation = Rotation(cv2.Rodrigues(r_vec)[0])
    translation = Translation(t_vec.ravel())

    model_view_t = rotation.compose_before(translation)
    proj_t = intrinsic_matrix.compose_before(flip_xy_yx()).compose_before(_drop_h)
    return model_view_t, proj_t


def solve_pnp_landmarks(
    image,
    mesh,
    group=None,
    intrinsic_matrix=None,
    distortion_coefficients=None,
    pnp_method=SOLVEPNP_ITERATIVE,
    n_iterations=100,
    reprojection_error=8.0,
    initial_transform=None,
):
    """
    Use OpenCV to solve the Perspective-N-Point problem (PnP). Uses Ransac PnP as this
    typically provides better results. The image and mesh must both have the same
    landmark group name attached.

    Note the intrinsic matrix (if given) must be in "OpenCV" space and thus
    has the "x" and "y" axes flipped w.r.t the menpo norm. E.g. the intrinsic matrix
    is defined as follows:

        [fx,  0, cx, 0]
        [ 0, fy, cy, 0]
        [ 0,  0,  1, 0]
        [ 0,  0,  0, 1]

    Parameters
    ----------
    image : :map`Image` or subclass
        The image to solve the PnP problem in. Must have landmarks attached.
    mesh : :map`Pointcloud` or subclass
        The mesh to solve the PnP problem with.  Must have landmarks attached.
    group : str, optional
        The name of the landmark group
    intrinsic_matrix : :map`Homogeneous`
        The intrinsic matrix - if not provided a basic pinhole camera is
        assumed.
    distortion_coefficients : ``(D,)`` `ndarray`
        The distortion coefficients (if not given assumes 0 coefficients). See the
        OpenCV documentation for the distortion coefficient types that are supported.
    pnp_method : int
        The OpenCV PNP method e.g. cv2.SOLVEPNP_ITERATIVE or otherwise
    n_iterations : int
        The number of iterations to perform
    reprojection_error : float
        The maximum reprojection error to allow for a point to be considered an
        inlier.
    initial_transform : :map`Homogeneous`
        The initialization for the cv2.SOLVEPNP_ITERATIVE method. Compatible
        with the returned model transformation returned by this method.

    Returns
    -------
    model_view_t : :map`Homogeneous`
        The combined ModelView transform. Can be used to place the mesh
        in "eye space".
    proj_t : :map`Homogeneous`
        A transform that can be used to project the input mesh landmarks
        back into the image

    Examples
    --------
    >>> from camera import pinhole_intrinsic_matrix
    >>> from menpo3d.correspond import solve_pnp
    >>> from menpo3d.rasterize.transform import (
    >>>      opengl_clip_matrix_from_mesh
    >>> )
    >>> template = m3dio.import_builtin_asset.template_ply()
    >>> image = mio.import_builtin_asset.lenna_png()

    >>> # Project the mesh back into the image
    >>> model_view_t, proj_t = solve_pnp(image, template, group="LJSON")
    >>> mesh_to_image = model_view_t.compose_before(proj_t)
    >>> image.landmarks['projected'] = mesh_to_image.apply(template.landmarks['LJSON'])

    >>> # Rasterize the mesh into the image
    >>> opengl_proj_t = opengl_clip_matrix_from_mesh(
    >>>     pinhole_intrinsic_matrix(image_height=image.height, image_width=image.width),
    >>>     image.width,
    >>>     image.height,
    >>>     view_t.apply(template),
    >>>     invert_y_axis=False,
    >>> )
    >>> rasterizer = GLRasterizer(height=image.height, width=image.width,
    >>>                           model_matrix=model_view_t.h_matrix,
    >>>                           projection_matrix=opengl_proj_t.h_matrix)
    >>> rasterized_image = rasterizer.rasterize_mesh(template)
    """
    if intrinsic_matrix is None:
        intrinsic_matrix = pinhole_intrinsic_matrix(image.shape[0], image.shape[1])

    return solve_pnp(
        image.landmarks[group],
        mesh.landmarks[group],
        intrinsic_matrix=intrinsic_matrix,
        distortion_coefficients=distortion_coefficients,
        pnp_method=pnp_method,
        n_iterations=n_iterations,
        reprojection_error=reprojection_error,
        initial_transform=initial_transform,
    )



from menpo.io.output.extensions import ljson_exporter
from .mesh import obj_exporter, ply_exporter


landmark_types = {".ljson": ljson_exporter}

mesh_types_buffer_support = {".obj": obj_exporter}

mesh_types_paths_only = {".ply": ply_exporter}

mesh_types = {}
mesh_types.update(mesh_types_buffer_support)
mesh_types.update(mesh_types_paths_only)



from functools import reduce

import numpy as np
from menpo.transform import Homogeneous, NonUniformScale, Scale, Translation


def gl_lookat_transform(invert_y_axis=True, invert_z_axis=True):
    """
    Equivalent to rotating y by 180, then rotating z by 180 if both invert_y_axis
    and invert_z_axis and True, then it simulates calling gluLookAt to
    look at the origin

        gluLookAt(0,0,0,0,0,1,0,-1,0)

    Parameters
    ----------
    invert_y_axis : bool
        If True, invert the Y-axis
    invert_z_axis : bool
        If True, invert the Z-axis

    Returns
    -------
    :map`Homogeneous`
        The matrix for applying the transformation
    """
    axes_flip_matrix = np.eye(4)
    axes_flip_matrix[1, 1] = -1 if invert_y_axis else 1
    axes_flip_matrix[2, 2] = -1 if invert_z_axis else 1
    return Homogeneous(axes_flip_matrix)


def flip_xy_yx():
    r"""
    Return a :map`Homogeneous` that flips the x and y axes.

    Returns
    -------
    :map`Homogeneous`
        The matrix for applying the transformation
    """
    return Homogeneous(
        np.array([[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
    )


def dims_3to2():
    r"""
    Return a :map`Homogeneous` that strips off the 3D axis of a 3D shape
    leaving just the first two axes.

    Returns
     ------
    :map`Homogeneous`
        The matrix for applying the transformation
    """
    return Homogeneous(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]]))


def dims_2to3(x=0):
    r"""
    Return a :map`Homogeneous` that adds on a 3rd axis to a 2D shape.

    Parameters
    ----------
    x : `float`, optional
        The value that will be assigned to the new third dimension

    Returns
     ------
    :map`Homogeneous`
        The matrix for applying the transformation
    """
    return Homogeneous(np.array([[1, 0, 0], [0, 1, 0], [0, 0, x], [0, 0, 1]]))


def model_to_clip_transform(points, xy_scale=0.9, z_scale=0.3):
    r"""
    Produces an Affine Transform which centres and scales 3D points to fit
    into the OpenGL clipping space ([-1, 1], [-1, 1], [1, 1-]). This can be
    used to construct an appropriate projection matrix for use in an
    orthographic Rasterizer. Note that the z-axis is flipped as is default in
    OpenGL - as a result this transform converts the right handed coordinate
    input into a left hand one.

    Parameters
    ----------
    points: :map:`PointCloud`
        The points that should be adjusted.
    xy_scale: `float` 0-1, optional
        Amount by which the boundary is relaxed so the points are not
        right against the edge. A value of 1 means the extremities of the
        point cloud will be mapped onto [-1, 1] [-1, 1] exactly (no boarder)
        A value of 0.5 means the points will be mapped into the range
        [-0.5, 0.5].

        Default: 0.9 (map to [-0.9, 0.9])

    z_scale: float 0-1, optional
        Scale factor by which the z-dimension is squeezed. A value of 1
        means the z-range of the points will be mapped to exactly fit in
        [1, -1]. A scale of 0.1 means the z-range is compressed to fit in the
        range [0.1, -0.1].

    Returns
    -------
    :map:`Affine`
        The affine transform that creates this mapping
    """
    # 1. Centre the points on the origin
    center = Translation(points.centre_of_bounds()).pseudoinverse()
    # 2. Scale the points to exactly fit the boundaries
    scale = Scale(points.range() / 2.0)
    # 3. Apply the relaxations requested - note the flip in the z axis!!
    # This is because OpenGL by default evaluates depth as bigger number ==
    # further away. Thus not only do we need to get to clip space [-1, 1] in
    # all dims) but we must invert the z axis so depth buffering is correctly
    # applied.
    b_scale = NonUniformScale([xy_scale, xy_scale, -z_scale])
    return center.compose_before(scale.pseudoinverse()).compose_before(b_scale)


def clip_to_image_transform(width, height):
    r"""
    Affine transform that converts 3D clip space coordinates into 2D image
    space coordinates. Note that the z axis of the clip space coordinates is
    ignored.

    Parameters
    ----------
    width: int
        The width of the image
    height: int
        The height of the image

    Returns
    -------
    :map`Homogeneous`
        A homogeneous transform that moves clip space coordinates into image
        space.
    """
    # 1. Remove the z axis from the clip space
    rem_z = dims_3to2()
    # 2. invert the y direction (up becomes down)
    invert_y = Scale([1, -1])
    # 3. [-1, 1] [-1, 1] -> [0, 2] [0, 2]
    t = Translation([1, 1])
    # 4. [0, 2] [0, 2] -> [0, 1] [0, 1]
    unit_scale = Scale(0.5, n_dims=2)
    # 5. [0, 1] [0, 1] -> [0, w - 1] [0, h - 1]
    im_scale = Scale([width - 1, height - 1])
    # 6. [0, w] [0, h] -> [0, h] [0, w]
    xy_yx = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]], dtype=np.float))
    # reduce the full transform chain to a single affine matrix
    transforms = [rem_z, invert_y, t, unit_scale, im_scale, xy_yx]
    return reduce(lambda a, b: a.compose_before(b), transforms)


def opengl_clip_matrix(
    intrinsic_matrix,
    image_width,
    image_height,
    near_plane,
    far_plane,
    invert_y_axis=True,
    invert_z_axis=True,
):
    """
    Build the transformation matrix that can project points into the OpenGL
    "clip space". Note that we take the convention following OpenGL and thus
    the given intrinsic matrix should be defined as follows:

            [fx,  0, cx, 0]
            [ 0, fy, cy, 0]
            [ 0,  0,  1, 0]
            [ 0,  0,  0, 1]

    Which is different than the convention normally followed in menpo where the axes
    always match the matrices e.g. this is likely "flipped" w.r.t the x and y axis
    in comparison to normal menpo conventions.

    Parameters
    ----------
    intrinsic_matrix : :map`Homogeneous`
        The intrinsic camera matrix
    image_width : int
        Image width
    image_height : int
        Image height
    near_plane : float
        Near plane (e.g. any vertices closer than this to the camera are culled from
        the view frustrum)
    far_plane :
        Far plane (e.g. any vertices further than this from the camera are culled from
        the view frustrum)
    invert_y_axis : bool
        If True, invert the Y-axis to account for the fact that OpenGL normally
        has the Y-axis flipped to a standard computer vision camera
    invert_z_axis : bool
        If True, invert the Z-axis to account for the fact that OpenGL normally
        has the Z-axis flipped to a standard computer vision camera

    Returns
    -------
    :map`Homogeneous`
        A homogeneous transform that can be used as a projection matrix in the
        rasterizer for transforming the vertices into OpenGL "clip space".
    """
    plane_prod = far_plane * near_plane
    far_sub_near = far_plane - near_plane

    fx = intrinsic_matrix.h_matrix[0, 0]
    fy = intrinsic_matrix.h_matrix[1, 1]
    cx = intrinsic_matrix.h_matrix[0, 2]
    cy = intrinsic_matrix.h_matrix[1, 2]

    # This assumes the window coordinate system has y pointing down
    clip_fx = 2 * fx / image_width
    clip_fy = -2 * fy / image_height
    clip_cx = (image_width - 2 * cx) / image_width
    clip_cy = (image_height - 2 * cy) / image_height

    clip_matrix = Homogeneous(
        np.array(
            [
                [clip_fx, 0, clip_cx, 0],
                [0, clip_fy, clip_cy, 0],
                [
                    0,
                    0,
                    (-far_plane - near_plane) / far_sub_near,
                    (-2.0 * plane_prod) / far_sub_near,
                ],
                [0, 0, -1, 0],
            ]
        )
    )

    lookat = gl_lookat_transform(
        invert_y_axis=invert_y_axis, invert_z_axis=invert_z_axis
    )

    return lookat.compose_before(clip_matrix)


def opengl_clip_matrix_from_mesh(
    intrinsic_matrix,
    image_width,
    image_height,
    mesh,
    invert_y_axis=True,
    invert_z_axis=True,
):
    r"""
    See documentation for ``opengl_clip_matrix`` for more information. This uses
    the mesh to define the near and far planes.

    Parameters
    ----------
    intrinsic_matrix : :map`Homogeneous`
        The intrinsic camera matrix
    image_width : int
        Image width
    image_height : int
        Image height
    mesh : :map`Pointcloud`
        Mesh to use for computing the near and far planes. Must be in "eye space" to
        ensure the planes are computed correctly. See
        `estimate_near_and_far_planes` for more information about "eye space".
    invert_y_axis : bool
        If True, invert the Y-axis to account for the fact that OpenGL normally
        has the Y-axis flipped to a standard computer vision camera
    invert_z_axis : bool
        If True, invert the Z-axis to account for the fact that OpenGL normally
        has the Z-axis flipped to a standard computer vision camera

    Returns
    -------
    :map`Homogeneous`
        A homogeneous transform that can be used as a projection matrix in the
        rasterizer for transforming the vertices into OpenGL "clip space".
    """
    near_plane, far_plane = estimate_near_and_far_planes(mesh)
    return opengl_clip_matrix(
        intrinsic_matrix,
        image_width=image_width,
        image_height=image_height,
        near_plane=near_plane,
        far_plane=far_plane,
        invert_y_axis=invert_y_axis,
        invert_z_axis=invert_z_axis,
    )


def estimate_near_and_far_planes(mesh):
    """
    Identify how far and near the mesh is in "eye space". We want to ensure that the
    near and far planes are set such that all of the mesh is displayed.

    We define the set of transformations as follows:

        1) The mesh initially exists in "object space"
        2) The Model transform converts from "object space" to "world space"
        3) The View transform converts from "world space" to "eye space"
        4) The Projection transform converts from "eye space" to "clip space"

    See http://www.songho.ca/opengl/gl_transform.html for more information.

    Parameters
    ----------
    mesh : :map:`PointCloud` or subclass
        The mesh in eye space.
    scale : float
        How much to scale the mesh range by to ensure that the mesh is within the
        near and far planes without worrying about numerical issues.

    Returns
    -------
    near_plane : float
        The near plane
    far_plane : float
        The far plane
    """
    near_bounds, far_bounds = mesh.bounds()
    z_near, z_far = near_bounds[-1], far_bounds[-1]
    # We assume that the mesh is already in eye space which means that it just
    # has to be converted into clip space. In this case the mesh cannot be
    # closer to the camera than 0 and the mesh is already as far away as it needs
    # to be
    return max(np.floor(z_near), 0), np.ceil(z_far)



import os
import sys
import warnings
from contextlib import contextmanager

import numpy as np
import scipy.sparse as sp
from io import UnsupportedOperation
from menpo.shape import TriMesh, PointCloud
from menpo.transform import Translation, UniformScale, AlignmentSimilarity
from menpo3d.morphablemodel.shapemodel import ShapeModel
from menpo3d.vtkutils import trimesh_to_vtk, VTKClosestPointLocator


@contextmanager
def stdout_redirected(to=os.devnull):
    r"""
    import os

    with stdout_redirected(to=filename):
        print("from Python")
        os.system("echo non-Python applications are also supported")
    """
    try:
        fd = sys.stdout.fileno()
    except UnsupportedOperation:
        # It's possible this is being run in an interpreter like an IPython
        # notebook where stdout doesn't behave the same as in a "normal" python
        # interpreter and in this case we cannot treat stdout like a file
        # descriptor
        warnings.warn(
            "Unable to duplicate stdout file descriptor, likely due "
            "to stdout having been replaced (e.g. a notebook)"
        )
        yield
    else:
        # assert that Python and C stdio write using the same file descriptor
        # assert libc.fileno(ctypes.c_void_p.in_dll(libc, "stdout")) == fd == 1

        def redirect_stdout(to):
            sys.stdout.close()  # + implicit flush()
            os.dup2(to.fileno(), fd)  # fd writes to 'to' file
            sys.stdout = os.fdopen(fd, "w")  # Python writes to fd

        with os.fdopen(os.dup(fd), "w") as old_stdout:
            with open(to, "w") as file:
                redirect_stdout(to=file)
            try:
                yield  # allow code to be run with the redirected stdout
            finally:
                # restore stdout.
                # buffering and flags such as CLOEXEC may be different
                redirect_stdout(to=old_stdout)


try:
    try:
        # First try the newer scikit-sparse namespace
        from sksparse.cholmod import cholesky_AAt
    except ImportError:
        # Fall back to the older scikits.sparse namespace
        from scikits.sparse.cholmod import cholesky_AAt

    # user has cholesky available - provide a fast solve
    def spsolve(sparse_X, dense_b):
        # wrap the cholesky call in a context manager that swallows the
        # low-level std-out to stop it from swamping our stdout (these low-level
        # prints come from METIS, but the solution behaves as normal)
        with stdout_redirected():
            factor = cholesky_AAt(sparse_X.T)
        return factor(sparse_X.T.dot(dense_b)).toarray()


except ImportError:
    # fallback to (much slower) scipy solve
    from scipy.sparse.linalg import spsolve as scipy_spsolve

    def spsolve(sparse_X, dense_b):
        warnings.warn(
            "suitesparse is not installed - NICP will run "
            "considerably (~5-10x) slower. If possible install "
            "suitesparse."
        )
        return scipy_spsolve(
            sparse_X.T.dot(sparse_X), sparse_X.T.dot(dense_b)
        ).toarray()


def node_arc_incidence_matrix(source):
    unique_edge_pairs = source.unique_edge_indices()
    m = unique_edge_pairs.shape[0]

    # Generate a "node-arc" (i.e. vertex-edge) incidence matrix.
    row = np.hstack((np.arange(m), np.arange(m)))
    col = unique_edge_pairs.T.ravel()
    data = np.hstack((-1 * np.ones(m), np.ones(m)))
    return sp.coo_matrix((data, (row, col))), unique_edge_pairs


def validate_weights(label, weights, n_points, n_iterations=None, verbose=False):
    if n_iterations is not None and len(weights) != n_iterations:
        raise ValueError(
            "Invalid {label}: - due to other weights there are "
            "{n_iterations} iterations but {n_weights} {label} "
            "were provided".format(
                label=label, n_iterations=n_iterations, n_weights=len(weights)
            )
        )
    invalid = []
    for i, weight in enumerate(weights):
        is_per_vertex = isinstance(weight, np.ndarray)
        if is_per_vertex and weight.shape != (n_points,):
            invalid.append("({}): {}".format(i, weight.shape[0]))

    if verbose and len(weights) >= 1:
        is_per_vertex = isinstance(weights[0], np.ndarray)
        if is_per_vertex:
            print("Using per-vertex {label}".format(label=label))
        else:
            print("Using global {label}".format(label=label))

    if len(invalid) != 0:
        raise ValueError(
            "Invalid {label}: expected shape ({n_points},) "
            "got: {invalid_cases}".format(
                label=label,
                n_points=n_points,
                invalid_cases="{}".format(", ".join(invalid)),
            )
        )


def non_rigid_icp(
    source,
    target,
    eps=1e-3,
    landmark_group=None,
    stiffness_weights=None,
    data_weights=None,
    landmark_weights=None,
    generate_instances=False,
    verbose=False,
):
    # call the generator version of NICP, always returning a generator
    generator = non_rigid_icp_generator(
        source,
        target,
        eps=eps,
        stiffness_weights=stiffness_weights,
        verbose=verbose,
        landmark_group=landmark_group,
        landmark_weights=landmark_weights,
        data_weights=data_weights,
    )
    # the handler decides whether the user get's details and each iteration
    # returned, or just the final result.
    return non_rigid_icp_generator_handler(generator, generate_instances)


def active_non_rigid_icp(
    model,
    target,
    eps=1e-3,
    stiffness_weights=None,
    data_weights=None,
    landmark_group=None,
    landmark_weights=None,
    model_mean_landmarks=None,
    generate_instances=False,
    verbose=False,
):
    model_mean = model.mean()

    if landmark_group is not None:

        # user better have provided model landmarks!
        if model_mean_landmarks is None:
            raise ValueError(
                "For Active NICP with landmarks the model_mean_landmarks "
                "need to be provided."
            )

        shape_model = ShapeModel(model)
        source_lms = model_mean_landmarks
        target_lms = target.landmarks[landmark_group]
        model_lms_index = model_mean.distance_to(source_lms).argmin(axis=0)
        shape_model_lms = shape_model.mask_points(model_lms_index)

        # Sim align the target lms to the mean before projecting
        target_lms_aligned = AlignmentSimilarity(target_lms, source_lms).apply(
            target_lms
        )

        # project to learn the weights for the landmark model
        weights = shape_model_lms.project(target_lms_aligned, n_components=20)

        # use these weights on the dense shape model to generate an improved
        # instance
        source = model.instance(weights)

        # update the source landmarks (for the alignment below)
        source.landmarks[landmark_group] = PointCloud(source.points[model_lms_index])
    else:
        # Start from the mean of the model
        source = model_mean

    # project onto the shape model to restrict the basis
    def project_onto_model(instance):
        return model.reconstruct(instance)

    # call the generator version of NICP, always returning a generator
    generator = non_rigid_icp_generator(
        source,
        target,
        eps=eps,
        stiffness_weights=stiffness_weights,
        data_weights=data_weights,
        landmark_weights=landmark_weights,
        landmark_group=landmark_group,
        v_i_update_func=project_onto_model,
        verbose=verbose,
    )
    # the handler decides whether the user get's details and each iteration
    # returned, or just the final result.
    return non_rigid_icp_generator_handler(generator, generate_instances)


def non_rigid_icp_generator_handler(generator, generate_instances):
    if generate_instances:
        # the user wants to inspect results per-iteration - return the iterator
        # directly to them
        return generator
    else:
        # the user is not interested in per-iteration results. Exhaust the
        # generator ourselves and return the last result only.
        while True:
            try:
                instance = next(generator)
            except StopIteration:
                return instance[0]


def non_rigid_icp_generator(
    source,
    target,
    eps=1e-3,
    stiffness_weights=None,
    data_weights=None,
    landmark_group=None,
    landmark_weights=None,
    v_i_update_func=None,
    verbose=False,
):
    r"""
    Deforms the source trimesh to align with to optimally the target.
    """
    # If landmarks are provided, we should always start with a simple
    # AlignmentSimilarity between the landmarks to initialize optimally.
    if landmark_group is not None:
        if verbose:
            print(
                "'{}' landmarks will be used as "
                "a landmark constraint.".format(landmark_group)
            )
            print("performing similarity alignment using landmarks")
        lm_align = AlignmentSimilarity(
            source.landmarks[landmark_group], target.landmarks[landmark_group]
        ).as_non_alignment()
        source = lm_align.apply(source)

    # Scale factors completely change the behavior of the algorithm - always
    # rescale the source down to a sensible size (so it fits inside box of
    # diagonal 1) and is centred on the origin. We'll undo this after the fit
    # so the user can use whatever scale they prefer.
    tr = Translation(-1 * source.centre())
    sc = UniformScale(1.0 / np.sqrt(np.sum(source.range() ** 2)), 3)
    prepare = tr.compose_before(sc)

    source = prepare.apply(source)
    target = prepare.apply(target)

    # store how to undo the similarity transform
    restore = prepare.pseudoinverse()

    n_dims = source.n_dims
    # Homogeneous dimension (1 extra for translation effects)
    h_dims = n_dims + 1
    points, trilist = source.points, source.trilist
    n = points.shape[0]  # record number of points

    edge_tris = source.boundary_tri_index()

    M_s, unique_edge_pairs = node_arc_incidence_matrix(source)

    # weight matrix
    G = np.identity(n_dims + 1)

    M_kron_G_s = sp.kron(M_s, G)

    # build octree for finding closest points on target.
    target_vtk = trimesh_to_vtk(target)
    closest_points_on_target = VTKClosestPointLocator(target_vtk)

    # save out the target normals. We need them for the weight matrix.
    target_tri_normals = target.tri_normals()

    # init transformation
    X_prev = np.tile(np.zeros((n_dims, h_dims)), n).T
    v_i = points

    if stiffness_weights is not None:
        if verbose:
            print("using user-defined stiffness_weights")
        validate_weights(
            "stiffness_weights", stiffness_weights, source.n_points, verbose=verbose
        )
    else:
        # these values have been empirically found to perform well for well
        # rigidly aligned facial meshes
        stiffness_weights = [50, 20, 5, 2, 0.8, 0.5, 0.35, 0.2]
        if verbose:
            print("using default " "stiffness_weights: {}".format(stiffness_weights))

    n_iterations = len(stiffness_weights)

    if landmark_weights is not None:
        if verbose:
            print("using user defined " "landmark_weights: {}".format(landmark_weights))
    elif landmark_group is not None:
        # these values have been empirically found to perform well for well
        # rigidly aligned facial meshes
        landmark_weights = [5, 2, 0.5, 0, 0, 0, 0, 0]
        if verbose:
            print("using default " "landmark_weights: {}".format(landmark_weights))
    else:
        # no landmark_weights provided - no landmark_group in use. We still
        # need a landmark group for the iterator
        landmark_weights = [None] * n_iterations

    # We should definitely have some landmark weights set now - check the
    # number is correct.
    # Note we say verbose=False, as we have done custom reporting above, and
    # per-vertex landmarks are not supported.
    validate_weights(
        "landmark_weights",
        landmark_weights,
        source.n_points,
        n_iterations=n_iterations,
        verbose=False,
    )

    if data_weights is not None:
        if verbose:
            print("using user-defined data_weights")
        validate_weights(
            "data_weights",
            data_weights,
            source.n_points,
            n_iterations=n_iterations,
            verbose=verbose,
        )
    else:
        data_weights = [None] * n_iterations
        if verbose:
            print("Not customising data_weights")

    # we need to prepare some indices for efficient construction of the D
    # sparse matrix.
    row = np.hstack(
        (np.repeat(np.arange(n)[:, None], n_dims, axis=1).ravel(), np.arange(n))
    )

    x = np.arange(n * h_dims).reshape((n, h_dims))
    col = np.hstack((x[:, :n_dims].ravel(), x[:, n_dims]))
    o = np.ones(n)

    if landmark_group is not None:
        source_lm_index = source.distance_to(source.landmarks[landmark_group]).argmin(
            axis=0
        )
        target_lms = target.landmarks[landmark_group]
        U_L = target_lms.points
        n_landmarks = target_lms.n_points
        lm_mask = np.in1d(row, source_lm_index)
        col_lm = col[lm_mask]
        # pull out the rows for the lms - but the values are
        # all wrong! need to map them back to the order of the landmarks
        row_lm_to_fix = row[lm_mask]
        source_lm_index_l = list(source_lm_index)
        row_lm = np.array([source_lm_index_l.index(r) for r in row_lm_to_fix])

    for i, (alpha, beta, gamma) in enumerate(
        zip(stiffness_weights, landmark_weights, data_weights), 1
    ):
        alpha_is_per_vertex = isinstance(alpha, np.ndarray)
        if alpha_is_per_vertex:
            # stiffness is provided per-vertex
            if alpha.shape[0] != source.n_points:
                raise ValueError()
            alpha_per_edge = alpha[unique_edge_pairs].mean(axis=1)
            alpha_M_s = sp.diags(alpha_per_edge).dot(M_s)
            alpha_M_kron_G_s = sp.kron(alpha_M_s, G)
        else:
            # stiffness is global - just a scalar multiply. Note that here
            # we don't have to recalculate M_kron_G_s
            alpha_M_kron_G_s = alpha * M_kron_G_s

        if verbose:
            a_str = (
                alpha
                if not alpha_is_per_vertex
                else "min: {:.2f}, max: {:.2f}".format(alpha.min(), alpha.max())
            )
            i_str = "{}/{}: stiffness: {}".format(i, len(stiffness_weights), a_str)
            if landmark_group is not None:
                i_str += "  lm_weight: {}".format(beta)
            print(i_str)

        j = 0
        while True:  # iterate until convergence
            j += 1  # track the iterations for this alpha/landmark weight

            # find nearest neighbour and the normals
            U, tri_indices = closest_points_on_target(v_i)

            # ---- WEIGHTS ----
            # 1.  Edges
            # Are any of the corresponding tris on the edge of the target?
            # Where they are we return a false weight (we *don't* want to
            # include these points in the solve)
            w_i_e = np.in1d(tri_indices, edge_tris, invert=True)

            # 2. Normals
            # Calculate the normals of the current v_i
            v_i_tm = TriMesh(v_i, trilist=trilist, copy=False)
            v_i_n = v_i_tm.vertex_normals()
            # Extract the corresponding normals from the target
            u_i_n = target_tri_normals[tri_indices]
            # If the dot of the normals is lt 0.9 don't contrib to deformation
            w_i_n = (u_i_n * v_i_n).sum(axis=1) > 0.9

            # 3. Self-intersection
            # This adds approximately 12% to the running cost and doesn't seem
            # to be very critical in helping mesh fitting performance so for
            # now it's removed. Revisit later.
            # # Build an intersector for the current deformed target
            # intersect = build_intersector(to_vtk(v_i_tm))
            # # budge the source points 1% closer to the target
            # source = v_i + ((U - v_i) * 0.5)
            # # if the vector from source to target intersects the deformed
            # # template we don't want to include it in the optimisation.
            # problematic = [i for i, (s, t) in enumerate(zip(source, U))
            #                if len(intersect(s, t)[0]) > 0]
            # print(len(problematic) * 1.0 / n)
            # w_i_i = np.ones(v_i_tm.n_points, dtype=np.bool)
            # w_i_i[problematic] = False

            # Form the overall w_i from the normals, edge case
            # for now disable the edge constraint (it was noisy anyway)
            w_i = w_i_n

            # w_i = np.logical_and(w_i_n, w_i_e).astype(np.float)

            # we could add self intersection at a later date too...
            # w_i = np.logical_and(np.logical_and(w_i_n,
            #                                     w_i_e,
            #                                     w_i_i).astype(np.float)

            prop_w_i = (n - w_i.sum() * 1.0) / n
            prop_w_i_n = (n - w_i_n.sum() * 1.0) / n
            prop_w_i_e = (n - w_i_e.sum() * 1.0) / n

            if gamma is not None:
                w_i = w_i * gamma

            # Build the sparse diagonal weight matrix
            W_s = sp.diags(w_i.astype(np.float)[None, :], [0])

            data = np.hstack((v_i.ravel(), o))
            D_s = sp.coo_matrix((data, (row, col)))

            to_stack_A = [alpha_M_kron_G_s, W_s.dot(D_s)]
            to_stack_B = [
                np.zeros((alpha_M_kron_G_s.shape[0], n_dims)),
                U * w_i[:, None],
            ]  # nullify nearest points by w_i

            if landmark_group is not None:
                D_L = sp.coo_matrix(
                    (data[lm_mask], (row_lm, col_lm)), shape=(n_landmarks, D_s.shape[1])
                )
                to_stack_A.append(beta * D_L)
                to_stack_B.append(beta * U_L)

            A_s = sp.vstack(to_stack_A).tocsr()
            B_s = sp.vstack(to_stack_B).tocsr()
            X = spsolve(A_s, B_s)

            # deform template
            v_i_prev = v_i
            v_i = D_s.dot(X)
            delta_v_i = v_i - v_i_prev

            if v_i_update_func:
                # custom logic is provided to update the current template
                # deformation. This is typically used by Active NICP.

                # take the v_i points matrix and convert back to a TriMesh in
                # the original space
                def_template = restore.apply(source.from_vector(v_i.ravel()))

                # perform the update
                updated_def_template = v_i_update_func(def_template)

                # convert back to points in the NICP space
                v_i = prepare.apply(updated_def_template.points)

            err = np.linalg.norm(X_prev - X, ord="fro")
            stop_criterion = err / np.sqrt(np.size(X_prev))

            if landmark_group is not None:
                src_lms = v_i[source_lm_index]
                lm_err = np.sqrt((src_lms - U_L) ** 2).sum(axis=1).mean()

            if verbose:
                v_str = (
                    " - {} stop crit: {:.3f}  "
                    "total: {:.0%}  norms: {:.0%}  "
                    "edges: {:.0%}".format(
                        j, stop_criterion, prop_w_i, prop_w_i_n, prop_w_i_e
                    )
                )
                if landmark_group is not None:
                    v_str += "  lm_err: {:.4f}".format(lm_err)

                print(v_str)

            X_prev = X

            # track the progress of the algorithm per-iteration
            info_dict = {
                "alpha": alpha,
                "iteration": j,
                "prop_omitted": prop_w_i,
                "prop_omitted_norms": prop_w_i_n,
                "prop_omitted_edges": prop_w_i_e,
                "delta": err,
                "mask_normals": w_i_n,
                "mask_edges": w_i_e,
                "mask_all": w_i,
                "nearest_points": restore.apply(U),
                "deformation_per_step": delta_v_i,
            }

            current_instance = source.copy()
            current_instance.points = v_i.copy()

            if landmark_group:
                info_dict["beta"] = beta
                info_dict["lm_err"] = lm_err
                current_instance.landmarks[landmark_group] = PointCloud(src_lms)

            yield restore.apply(current_instance), info_dict

            if stop_criterion < eps:
                break



import numpy as np
from menpo.base import Vectorizable
from menpo.transform import Homogeneous, Rotation, Transform, Translation


class OrthographicProjection(Transform, Vectorizable):
    def __init__(self, focal_length, image_shape):
        self.focal_length = focal_length
        self.height = image_shape[0]
        self.width = image_shape[1]

    @property
    def n_dims(self):
        return 3

    @property
    def n_parameters(self):
        return 1

    def as_vector(self):
        return np.array([self.focal_length])

    def _from_vector_inplace(self, vector):
        self.focal_length = vector[0]

    def _apply(self, x, **kwargs):
        f = self.focal_length
        c_x = self.width / 2
        c_y = self.height / 2

        output = np.empty_like(x)
        output[:, 0] = (f * x[:, 1]) + c_y
        output[:, 1] = (f * x[:, 0]) + c_x
        output[:, 2] = x[:, 2]

        return output


class PerspectiveProjection(OrthographicProjection):
    def _apply(self, x, **kwargs):
        f = self.focal_length
        c_x = self.width / 2
        c_y = self.height / 2

        output = np.empty_like(x)
        output[:, 0] = (f * x[:, 1]) / x[:, 2] + c_y
        output[:, 1] = (f * x[:, 0]) / x[:, 2] + c_x
        output[:, 2] = x[:, 2]

        return output

    @classmethod
    def init_from_2d_projected_shape(
        cls,
        points_3d,
        points_image,
        image_shape,
        focal_length=None,
        distortion_coeffs=None,
    ):
        from menpo3d.correspond import solve_pnp  # Avoid circular import

        model_view_t, _ = solve_pnp(
            points_image,
            points_3d,
            pinhole_intrinsic_matrix(
                image_shape[0], image_shape[1], focal_length=focal_length
            ),
            distortion_coefficients=distortion_coeffs,
        )
        rotation = Rotation(model_view_t.h_matrix[:3, :3])
        translation = Translation(model_view_t.h_matrix[:3, -1])
        return OrthographicCamera(
            rotation, translation, OrthographicProjection(focal_length, image_shape)
        )


class OrthographicCamera(Vectorizable):
    def __init__(self, rotation, translation, projection):
        self.rotation_transform = rotation
        self.translation_transform = translation
        self.projection_transform = projection

    @property
    def focal_length(self):
        return self.projection_transform.focal_length

    @classmethod
    def init_from_image_shape_and_vector(cls, image_shape, vector):
        r = Rotation.init_identity(n_dims=3)
        t = Translation.init_identity(n_dims=3)
        p = OrthographicProjection(focal_length=1, image_shape=image_shape)
        return cls(r, t, p).from_vector(vector)

    @classmethod
    def init_from_2d_projected_shape(
        cls,
        points_3d,
        points_image,
        image_shape,
        focal_length=None,
        distortion_coeffs=None,
    ):
        raise NotImplementedError(
            "Orthographic camera pose estimation not " "implemented."
        )

    def apply(self, instance, **kwargs):
        return self.camera_transform.apply(instance)

    @property
    def n_parameters(self):
        return (
            self.projection_transform.n_parameters
            + self.rotation_transform.n_parameters
            + self.translation_transform.n_parameters
        )

    def as_vector(self):
        # focal_length, q_w, q_x, q_y, q_z, t_x, t_y, t_z
        params = np.zeros(self.n_parameters)

        # focal length
        params[:1] = self.projection_transform.as_vector()

        # 4 parameters: q_w, q_x, q_y, q_z
        params[1:5] = self.rotation_transform.as_vector()

        # 3 parameters: t_x, t_y, t_z
        params[5:] = self.translation_transform.as_vector()
        return params

    def _from_vector_inplace(self, vector):
        self.projection_transform._from_vector_inplace(vector[:1])
        self.rotation_transform._from_vector_inplace(vector[1:5])
        self.translation_transform._from_vector_inplace(vector[5:])

    @property
    def view_transform(self):
        return self.rotation_transform.compose_before(self.translation_transform)

    @property
    def camera_transform(self):
        return self.view_transform.compose_before(self.projection_transform)


class PerspectiveCamera(OrthographicCamera):
    @classmethod
    def init_from_image_shape_and_vector(cls, image_shape, vector):
        r = Rotation.init_identity(n_dims=3)
        t = Translation.init_identity(n_dims=3)
        p = PerspectiveProjection(focal_length=1, image_shape=image_shape)
        return cls(r, t, p).from_vector(vector)

    @classmethod
    def init_from_2d_projected_shape(
        cls,
        points_3d,
        points_image,
        image_shape,
        focal_length=None,
        distortion_coeffs=None,
    ):
        from menpo3d.correspond import solve_pnp  # Avoid circular import

        model_view_t, _ = solve_pnp(
            points_image,
            points_3d,
            pinhole_intrinsic_matrix(
                image_shape[0], image_shape[1], focal_length=focal_length
            ),
            distortion_coefficients=distortion_coeffs,
        )
        rotation = Rotation(model_view_t.h_matrix[:3, :3])
        translation = Translation(model_view_t.h_matrix[:3, -1])
        return PerspectiveCamera(
            rotation, translation, PerspectiveProjection(focal_length, image_shape)
        )


def pinhole_intrinsic_matrix(image_height, image_width, focal_length=None):
    r"""
    Create a basic "pinhole" type camera intrinsic matrix. Focal length is in pixels
    and principal point is in the image center. Note this follows OpenCV image
    conventions and thus the "first" axis is the x-axis rather than the typical
    menpo convention of the "first" axis being the y-axis.

        [fx,  0, cx, 0]
        [ 0, fy, cy, 0]
        [ 0,  0,  1, 0]
        [ 0,  0,  0, 1]

    Parameters
    ----------
    image_height : int
        Image height
    image_width : int
        Image width
    focal_length : float, optional
        If given, the focal length (fx=fy) in pixels. If not given, the max
        of the width and height is used.

    Returns
    -------
    :map`Homogeneous`
        3D camera intrinsics matrix as a Homogeneous matrix
    """
    if focal_length is None:
        focal_length = max(image_height, image_width)
    return Homogeneous(
        np.array(
            [
                [focal_length, 0, image_width / 2, 0],
                [0, focal_length, image_height / 2, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ]
        )
    )



from menpo.io.input.landmark import ljson_importer
from .mesh import wrl_importer, mjson_importer, obj_importer, stl_importer, ply_importer
from .landmark_mesh import bnd_importer, lan_importer, lm3_importer, pts_mesh_importer
from .lsfm import lsfm_model_importer

# TODO: Add PLY (ASCII and binary) and OFF importers
mesh_types = {
    ".obj": obj_importer,
    ".stl": stl_importer,
    ".ply": ply_importer,
    ".wrl": wrl_importer,
    ".mjson": mjson_importer,
}

mesh_landmark_types = {
    ".pts3": pts_mesh_importer,
    ".lm3": lm3_importer,
    ".lan": lan_importer,
    ".bnd": bnd_importer,
    ".ljson": ljson_importer,
}

lsfm_types = {".mat": lsfm_model_importer}



from .input import (
    import_builtin_asset,
    mesh_paths,
    import_landmark_file,
    import_landmark_files,
    data_path_to,
    data_dir_path,
    ls_builtin_assets,
    landmark_file_paths,
    import_mesh,
    import_meshes,
    register_landmark_importer,
    register_mesh_importer,
    import_lsfm_model,
)
from .output import export_landmark_file, export_mesh, export_textured_mesh



import numpy as np

from menpo.shape import PointCloud
from menpo.transform.base import Transformable


class ShapeModel(Transformable):
    r"""
    Wrapper around a :map:`PCAModel` that makes it transformable and
    maskable in both dimensions and points.
    """

    def __init__(self, model):
        self.model = model

    @property
    def n_dims(self):
        return self.model.template_instance.n_dims

    @property
    def n_components(self):
        return self.model.n_components

    @property
    def _c(self):
        return self.model._components.reshape([self.n_components, -1, self.n_dims])

    def mask_points(self, mask):
        new_model = self.model.copy()
        new_model._components = self._c[:, mask].reshape([self.n_components, -1])
        mean = self.model.mean().points[mask]
        new_model._mean = mean.ravel()
        new_model.template_instance = PointCloud(mean)
        return ShapeModel(new_model)

    def mask_dims(self, mask):
        new_model = self.model.copy()
        new_model._components = self._c[..., mask].reshape([self.n_components, -1])
        mean = self.model.mean().points[:, mask]
        new_model._mean = mean.ravel()
        new_model.template_instance = PointCloud(mean)
        return ShapeModel(new_model)

    def _transform(self, transform):
        new_model = self.model.copy()
        new_model._components = np.concatenate(
            [transform(c)[None] for c in self._c]
        ).reshape([self.n_components, -1])
        new_model._mean = transform(self.model.mean().points).ravel()
        return ShapeModel(new_model)

    def project(self, instance, n_components=None):
        if n_components is not None:
            prev_n_components = self.model.n_active_components
            self.model.n_active_components = n_components
        weights = np.linalg.lstsq(
            self.model.components.T, instance.as_vector() - self.model.mean_vector
        )[0]
        if n_components is not None:
            self.model.n_active_components = prev_n_components
        return weights



import numpy as np

from menpo3d.camera import PerspectiveCamera, OrthographicCamera


def d_perspective_camera_d_shape_parameters(shape_pc_uv, warped_uv, camera):
    """
    Calculates the derivative of the perspective projection with respect to the
    shape parameters.

    Parameters
    ----------
    shape_pc_uv : ``(n_points, 3, n_parameters)`` `ndarray`
        The (sampled) basis of the shape model.
    warped_uv : ``(n_points, 3)`` `ndarray`
        The shape instance with the view transform (rotation and translation)
        applied on it.
    camera : `menpo3d.camera.PerspectiveCamera`
        The camera object that is responsible of projecting the model to the
        image plane.

    Returns
    -------
    dw_da : ``(2, n_shape_parameters, n_points)`` `ndarray`
        The derivative of the perspective camera transform with respect to
        the shape parameters.
    """
    n_points, n_dims, n_parameters = shape_pc_uv.shape
    assert n_dims == 3

    # Compute constant
    # (focal length divided by squared Z dimension of warped shape)
    z = warped_uv[:, 2]

    # n_dims, n_parameters, n_points
    dw_da = camera.rotation_transform.apply(shape_pc_uv.transpose(0, 2, 1)).T

    dw_da[:2] -= warped_uv[:, :2].T[:, None] * dw_da[2] / z

    return camera.projection_transform.focal_length * dw_da[:2] / z


def d_orthographic_camera_d_shape_parameters(shape_pc_uv, camera):
    """
    Calculates the derivative of the orthographic projection with respect to
    the shape parameters.

    Parameters
    ----------
    shape_pc_uv : ``(n_points, 3, n_parameters)`` `ndarray`
        The (sampled) basis of the shape model.
    camera : `menpo3d.camera.OrthographicCamera`
        The camera object that is responsible of projecting the model to the
        image plane.

    Returns
    -------
    dw_da : ``(2, n_shape_parameters, n_points)`` `ndarray`
        The derivative of the perspective camera transform with respect to
        the shape parameters.
    """
    n_points, n_dims, n_parameters = shape_pc_uv.shape
    assert n_dims == 3

    # n_dims, n_parameters, n_points
    dp_da_uv = camera.rotation_transform.apply(shape_pc_uv.transpose(0, 2, 1)).T

    return camera.projection_transform.focal_length * dp_da_uv[:2]


def d_camera_d_shape_parameters(camera, warped_uv, shape_pc_uv):
    if camera.__class__ == PerspectiveCamera:
        dp_da = d_perspective_camera_d_shape_parameters(shape_pc_uv, warped_uv, camera)
    elif camera.__class__ == OrthographicCamera:
        dp_da = d_orthographic_camera_d_shape_parameters(shape_pc_uv, camera)
    else:
        raise ValueError("Camera must be either Perspective or " "Orthographic.")
    return dp_da


def d_perspective_camera_d_camera_parameters(
    warped_uv, camera, with_focal_length=True, with_first_quaternion=False
):
    """
    Calculates the derivative of the perspective projection with respect to the
    camera parameters.

    Parameters
    ----------
    warped_uv : ``(n_points, 3)`` `ndarray`
        The shape instance with the view transform (rotation and translation)
        applied on it.
    camera : `menpo3d.camera.PerspectiveCamera`
        The camera object that is responsible of projecting the model to the
        image plane.
    with_focal_length : `bool`, optional
        If ``False``, then the derivative with respect to the focal length
        parameter won't be returned.
    with_first_quaternion : `bool`, optional
        If ``False``, then the derivative with respect to the first
        quaternion parameter won't be returned.

    Returns
    -------
    dw_dr : ``(2, n_parameters, n_points)`` `ndarray`
        The derivative of the perspective camera transform with respect to
        the camera parameters.
    """
    n_points, n_dims = warped_uv.shape
    assert n_dims == 3

    # Find total number of parameters
    n_parameters = camera.n_parameters
    if not with_focal_length:
        n_parameters -= 1
    if not with_first_quaternion:
        n_parameters -= 1

    # Initialize derivative
    dw_dr = np.zeros((2, n_parameters, n_points))

    # Initialize parameter counter
    r = 0

    # Get z-component of warped
    z = warped_uv[:, 2]

    # Focal length, if requested
    if with_focal_length:
        dw_dr[:, r] = (warped_uv[:, :2] / z[..., None]).T
        r += 1

    # Quaternions
    centered_warped_uv = camera.translation_transform.pseudoinverse().apply(warped_uv).T
    # q_1, if requested
    if with_first_quaternion:
        r0 = 2 * np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]).dot(centered_warped_uv).T
        dw_dr[:, r] = r0[:, :2].T - r0[:, 2] * warped_uv[:, :2].T / z
        r += 1

    # q_2
    r1 = 2 * np.array([[0, 0, 0], [0, 0, -1], [0, 1, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = r1[:, :2].T - r1[:, 2] * warped_uv[:, :2].T / z
    r += 1

    # q_3
    r2 = 2 * np.array([[0, 0, 1], [0, 0, 0], [-1, 0, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = r2[:, :2].T - r2[:, 2] * warped_uv[:, :2].T / z
    r += 1

    # q_4
    r3 = 2 * np.array([[0, -1, 0], [1, 0, 0], [0, 0, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = r3[:, :2].T - r3[:, 2] * warped_uv[:, :2].T / z
    r += 1

    # constant multiplication
    i1 = 1 if with_focal_length else 0
    i2 = i1 + 4 if with_first_quaternion else i1 + 3
    dw_dr[:, i1:i2] *= camera.projection_transform.focal_length / z

    # Translations
    # t_x
    dw_dr[0, r] = camera.projection_transform.focal_length / z
    r += 1
    # t_y
    dw_dr[1, r] = camera.projection_transform.focal_length / z
    r += 1
    # t_z
    dw_dr[:, r] = (
        -camera.projection_transform.focal_length * warped_uv[:, :2] / z[..., None] ** 2
    ).T

    return dw_dr


def d_orthographic_camera_d_camera_parameters(
    warped_uv, camera, with_focal_length=True, with_first_quaternion=False
):
    """
    Calculates the derivative of the orthographic projection with respect to the
    camera parameters.

    Parameters
    ----------
    warped_uv : ``(n_points, 3)`` `ndarray`
        The shape instance with the view transform (rotation and translation)
        applied on it.
    camera : `menpo3d.camera.PerspectiveCamera`
        The camera object that is responsible of projecting the model to the
        image plane.
    with_focal_length : `bool`, optional
        If ``False``, then the derivative with respect to the focal length
        parameter won't be returned.
    with_first_quaternion : `bool`, optional
        If ``False``, then the derivative with respect to the first
        quaternion parameter won't be returned.

    Returns
    -------
    dw_dr : ``(2, n_parameters, n_points)`` `ndarray`
        The derivative of the orthographic camera transform with respect to
        the camera parameters.
    """
    n_points, n_dims = warped_uv.shape
    assert n_dims == 3

    # Find total number of parameters
    n_parameters = camera.n_parameters
    if not with_focal_length:
        n_parameters -= 1
    if not with_first_quaternion:
        n_parameters -= 1

    # Initialize derivative
    dw_dr = np.zeros((2, n_parameters, n_points))

    # Initialize parameter counter
    r = 0

    # Focal length, if requested
    if with_focal_length:
        dw_dr[:, r] = warped_uv[:, :2].T
        r += 1

    # Quaternions
    centered_warped_uv = camera.translation_transform.pseudoinverse().apply(warped_uv).T

    # q_1, if requested
    if with_first_quaternion:
        r0 = 2 * np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]).dot(centered_warped_uv).T
        dw_dr[:, r] = camera.projection_transform.focal_length * r0[:, :2].T
        r += 1

    # q_2
    r1 = 2 * np.array([[0, 0, 0], [0, 0, -1], [0, 1, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = camera.projection_transform.focal_length * r1[:, :2].T
    r += 1

    # q_3
    r2 = 2 * np.array([[0, 0, 1], [0, 0, 0], [-1, 0, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = camera.projection_transform.focal_length * r2[:, :2].T
    r += 1

    # q_4
    r3 = 2 * np.array([[0, -1, 0], [1, 0, 0], [0, 0, 0]]).dot(centered_warped_uv).T
    dw_dr[:, r] = camera.projection_transform.focal_length * r3[:, :2].T
    r += 1

    # Translations
    # t_x
    dw_dr[0, r] = camera.projection_transform.focal_length
    r += 1

    # t_y
    dw_dr[1, r] = camera.projection_transform.focal_length
    r += 1

    return dw_dr


def d_camera_d_camera_parameters(camera, warped_uv, with_focal_length):
    if camera.__class__ == PerspectiveCamera:
        dp_dr = d_perspective_camera_d_camera_parameters(
            warped_uv,
            camera,
            with_focal_length=with_focal_length,
            with_first_quaternion=False,
        )
    elif camera.__class__ == OrthographicCamera:
        dp_dr = d_orthographic_camera_d_camera_parameters(
            warped_uv,
            camera,
            with_focal_length=with_focal_length,
            with_first_quaternion=False,
        )
    else:
        raise ValueError("Camera must be either Perspective or " "Orthographic.")
    return dp_dr



import numpy as np

from menpo.feature import gradient as fast_gradient
from menpo.image import Image
from menpo.visualize import print_dynamic

from menpo3d.rasterize import rasterize_barycentric_coordinates

from .derivatives import d_camera_d_camera_parameters, d_camera_d_shape_parameters
from ..result import MMAlgorithmResult


class LucasKanade(object):
    def __init__(self, model, n_samples, eps=1e-3):
        self.model = model
        self.eps = eps
        self.n_samples = n_samples
        # Call precompute
        self._precompute()

    @property
    def n(self):
        r"""
        Returns the number of active components of the shape model.

        :type: `int`
        """
        return self.model.shape_model.n_active_components

    @property
    def m(self):
        r"""
        Returns the number of active components of the texture model.

        :type: `int`
        """
        return self.model.texture_model.n_active_components

    @property
    def n_vertices(self):
        r"""
        Returns the number of vertices of the shape model's trimesh.

        :type: `int`
        """
        return self.model.shape_model.template_instance.n_points

    @property
    def n_channels(self):
        r"""
        Returns the number of channels of the texture model.

        :type: `int`
        """
        return self.model.n_channels

    def visible_sample_points(self, instance_in_img, image_shape):
        # Inverse rendering
        yx, bcoords, tri_indices = rasterize_barycentric_coordinates(
            instance_in_img, image_shape
        )

        # Select triangles randomly
        rand = np.random.permutation(bcoords.shape[0])
        bcoords = bcoords[rand[: self.n_samples]]
        yx = yx[rand[: self.n_samples]]
        tri_indices = tri_indices[rand[: self.n_samples]]

        # Build the vertex indices (3 per pixel) for the visible triangles
        vertex_indices = instance_in_img.trilist[tri_indices]

        return vertex_indices, bcoords, tri_indices, yx

    def sample(self, x, bcoords, vertex_indices):
        per_vert_per_pixel = x[vertex_indices]
        return np.sum(per_vert_per_pixel * bcoords[..., None], axis=1)

    def gradient(self, image):
        # Compute the gradient of the image
        grad = fast_gradient(image)

        # Create gradient image for X and Y
        grad_y = Image(grad.pixels[: self.n_channels])
        grad_x = Image(grad.pixels[self.n_channels :])

        return grad_x, grad_y

    def compute_cost(
        self,
        data_error,
        lms_error,
        shape_parameters,
        texture_parameters,
        shape_prior_weight,
        texture_prior_weight,
        landmarks_prior_weight,
    ):
        # Cost of data term
        current_cost = data_error.T.dot(data_error)

        # Cost of shape prior
        if shape_prior_weight is not None:
            current_cost += shape_prior_weight * np.sum(
                (shape_parameters ** 2) * self.J_shape_prior
            )

        # Cost of texture prior
        if texture_prior_weight is not None:
            current_cost += texture_prior_weight * np.sum(
                (texture_parameters ** 2) * self.J_texture_prior
            )

        # Cost of landmarks prior
        if landmarks_prior_weight is not None:
            current_cost += landmarks_prior_weight * lms_error.T.dot(lms_error)

        return current_cost

    def J_lms(self, camera, warped_uv, shape_pc_uv, camera_update, focal_length_update):
        # Compute derivative of camera wrt shape and camera parameters
        J = d_camera_d_shape_parameters(camera, warped_uv, shape_pc_uv)
        n_camera_parameters = 0
        if camera_update:
            dp_dr = d_camera_d_camera_parameters(
                camera, warped_uv, with_focal_length=focal_length_update
            )
            J = np.hstack((J, dp_dr))
            n_camera_parameters = dp_dr.shape[1]

        # Reshape to : n_params x (2 * N)
        n_params = J.shape[1]
        J = np.transpose(J, (1, 0, 2)).reshape(n_params, -1)
        return J, n_camera_parameters

    def _precompute(self):
        # Rescale shape and appearance components to have size:
        # n_vertices x (n_active_components * n_dims)
        shape_pc = self.model.shape_model.components.T
        self.shape_pc = shape_pc.reshape([self.n_vertices, -1])

        # Priors
        self.J_shape_prior = 1.0 / np.array(self.model.shape_model.eigenvalues)
        self.J_texture_prior = 1.0 / np.array(self.model.texture_model.eigenvalues)
        self.shape_pc_lms = shape_pc.reshape([self.n_vertices, 3, -1])[
            self.model.model_landmarks_index
        ]


class SimultaneousForwardAdditive(LucasKanade):
    r"""
    Class for defining Simultaneous Forward Additive Morphable Model
    optimization algorithm.
    """

    def run(
        self,
        image,
        initial_mesh,
        camera,
        gt_mesh=None,
        max_iters=20,
        camera_update=False,
        focal_length_update=False,
        reconstruction_weight=1.0,
        shape_prior_weight=1.0,
        texture_prior_weight=1.0,
        landmarks=None,
        landmarks_prior_weight=1.0,
        return_costs=False,
        verbose=True,
    ):
        # Parse landmarks prior options
        if landmarks is None or landmarks_prior_weight is None:
            landmarks_prior_weight = None
            landmarks = None
        lms_points = None
        if landmarks is not None:
            lms_points = landmarks.points[:, [1, 0]]

        # Retrieve camera parameters from the provided camera object.
        # Project provided instance to retrieve shape and texture parameters.
        camera_parameters = camera.as_vector()
        shape_parameters = self.model.shape_model.project(initial_mesh)
        texture_parameters = self.model.project_instance_on_texture_model(initial_mesh)

        # Reconstruct provided instance
        instance = self.model.instance(
            shape_weights=shape_parameters, texture_weights=texture_parameters
        )

        # Compute input image gradient
        grad_x, grad_y = self.gradient(image)

        # Initialize lists
        shape_parameters_per_iter = [shape_parameters]
        texture_parameters_per_iter = [texture_parameters]
        camera_per_iter = [camera]
        instance_per_iter = [instance.rescale_texture(0.0, 1.0)]
        costs = None
        if return_costs:
            costs = []

        # Initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Main loop
        while k < max_iters and eps > self.eps:
            if verbose:
                print_dynamic("{}/{}".format(k + 1, max_iters))
            # Apply camera projection on current instance
            instance_in_image = camera.apply(instance)

            # Compute indices locations for sampling
            (vertex_indices, bcoords, tri_indices, yx) = self.visible_sample_points(
                instance_in_image, image.shape
            )

            # Warp the mesh with the view matrix (rotation + translation)
            instance_w = camera.view_transform.apply(instance.points)

            # Sample all the terms from the model part at the sample locations
            warped_uv = self.sample(instance_w, bcoords, vertex_indices)
            texture_uv = instance.sample_texture_with_barycentric_coordinates(
                bcoords, tri_indices
            )
            texture_pc_uv = self.model.sample_texture_model(bcoords, tri_indices)
            shape_pc_uv = self.sample(self.shape_pc, bcoords, vertex_indices)
            # Reshape shape basis after sampling
            shape_pc_uv = shape_pc_uv.reshape([self.n_samples, 3, -1])

            # Sample all the terms from the image part at the sample locations
            img_uv = image.sample(yx)
            grad_x_uv = grad_x.sample(yx)
            grad_y_uv = grad_y.sample(yx)

            # Compute error
            img_error_uv = (img_uv - texture_uv.T).ravel()

            # Compute Jacobian, SD and Hessian of data term
            if reconstruction_weight is not None:
                sd, n_camera_parameters = self.J_data(
                    camera,
                    warped_uv,
                    shape_pc_uv,
                    texture_pc_uv,
                    grad_x_uv,
                    grad_y_uv,
                    camera_update,
                    focal_length_update,
                    reconstruction_weight,
                )
                hessian = sd.dot(sd.T)
                sd_error = sd.dot(img_error_uv)
            else:
                n_camera_parameters = 0
                if camera_update:
                    if focal_length_update:
                        n_camera_parameters = camera.n_parameters - 1
                    else:
                        n_camera_parameters = camera.n_parameters - 2
                hessian = np.zeros(
                    (self.n + n_camera_parameters, self.n + n_camera_parameters)
                )
                sd_error = np.zeros(self.n + n_camera_parameters)

            # Compute Jacobian, update SD and Hessian wrt shape prior
            if shape_prior_weight is not None:
                sd_shape = shape_prior_weight * self.J_shape_prior
                hessian[: self.n, : self.n] += np.diag(sd_shape)
                sd_error[: self.n] += sd_shape * shape_parameters

            # Compute Jacobian, update SD and Hessian wrt texture prior
            if texture_prior_weight is not None:
                idx = self.n + n_camera_parameters
                sd_texture = texture_prior_weight * self.J_texture_prior
                hessian[idx:, idx:] += np.diag(sd_texture)
                sd_error[idx:] += sd_texture * texture_parameters

            # Compute Jacobian, update SD and Hessian wrt landmarks prior
            lms_error = None
            if landmarks_prior_weight is not None:
                # Get projected instance on landmarks and error term
                warped_lms = instance_in_image.points[self.model.model_landmarks_index]
                lms_error = (warped_lms[:, [1, 0]] - lms_points).T.ravel()
                warped_view_lms = instance_w[self.model.model_landmarks_index]

                # Jacobian and Hessian wrt shape parameters
                sd_lms, n_camera_parameters = self.J_lms(
                    camera,
                    warped_view_lms,
                    self.shape_pc_lms,
                    camera_update,
                    focal_length_update,
                )
                idx = self.n + n_camera_parameters
                hessian[:idx, :idx] += landmarks_prior_weight * sd_lms.dot(sd_lms.T)
                sd_error[:idx] = landmarks_prior_weight * sd_lms.dot(lms_error)

            if return_costs:
                costs.append(
                    self.compute_cost(
                        img_error_uv,
                        lms_error,
                        shape_parameters,
                        texture_parameters,
                        shape_prior_weight,
                        texture_prior_weight,
                        landmarks_prior_weight,
                    )
                )

            # Solve to find the increment of parameters
            d_shape, d_camera, d_texture = self.solve(
                hessian,
                sd_error,
                reconstruction_weight,
                camera_update,
                focal_length_update,
                camera,
            )

            # Update parameters
            shape_parameters += d_shape
            if camera_update:
                camera_parameters = camera_parameters_update(
                    camera_parameters, d_camera
                )
                camera = camera.from_vector(camera_parameters)
            texture_parameters += d_texture

            # Generate the updated instance
            instance = self.model.instance(
                shape_weights=shape_parameters, texture_weights=texture_parameters
            )

            # Update lists
            shape_parameters_per_iter.append(shape_parameters)
            texture_parameters_per_iter.append(texture_parameters)
            camera_per_iter.append(camera)
            instance_per_iter.append(instance.rescale_texture(0.0, 1.0))

            # Increase iteration counter
            k += 1

            # shape_parameters, texture_parameters, camera_parameters = yield \
            #     shape_parameters, texture_parameters, camera_parameters

        return MMAlgorithmResult(
            shape_parameters=shape_parameters_per_iter,
            texture_parameters=texture_parameters_per_iter,
            meshes=instance_per_iter,
            camera_transforms=camera_per_iter,
            image=image,
            initial_mesh=initial_mesh.rescale_texture(0.0, 1.0),
            initial_camera_transform=camera_per_iter[0],
            gt_mesh=gt_mesh,
            costs=costs,
        )

    def solve(
        self,
        hessian,
        sd_error,
        reconstruction_prior_weight,
        camera_update,
        focal_length_update,
        camera,
    ):
        # Solve
        ds = -np.linalg.solve(hessian, sd_error)

        # Get shape parameters increment
        d_shape = ds[: self.n]

        # Initialize texture parameters update
        d_texture = np.zeros(self.m)

        # Get camera parameters increment
        if camera_update:
            # Keep the rest
            ds = ds[self.n :]

            # If focal length is not updated, then set its increment to zero
            if not focal_length_update:
                ds = np.insert(ds, 0, [0.0])

            # Set increment of the 1st quaternion to one
            ds = np.insert(ds, 1, [1.0])

            # Get camera parameters update
            d_camera = ds[: camera.n_parameters]

            # Get texture parameters increment
            if reconstruction_prior_weight is not None:
                d_texture = ds[camera.n_parameters :]
        else:
            d_camera = None
            if reconstruction_prior_weight is not None:
                d_texture = ds[self.n :]

        return d_shape, d_camera, d_texture

    def J_data(
        self,
        camera,
        warped_uv,
        shape_pc_uv,
        texture_pc_uv,
        grad_x_uv,
        grad_y_uv,
        camera_update,
        focal_length_update,
        reconstruction_prior_weight,
    ):
        # Compute derivative of camera wrt shape and camera parameters
        dp_da_dr = d_camera_d_shape_parameters(camera, warped_uv, shape_pc_uv)
        n_camera_parameters = 0
        if camera_update:
            dp_dr = d_camera_d_camera_parameters(
                camera, warped_uv, with_focal_length=focal_length_update
            )
            dp_da_dr = np.hstack((dp_da_dr, dp_dr))
            n_camera_parameters = dp_dr.shape[1]

        # Multiply image gradient with camera derivative
        permuted_grad_x = np.transpose(grad_x_uv[..., None], (0, 2, 1))
        permuted_grad_y = np.transpose(grad_y_uv[..., None], (0, 2, 1))
        J = permuted_grad_x * dp_da_dr[0] + permuted_grad_y * dp_da_dr[1]

        # Computer derivative of texture wrt texture parameters
        dt_db = -np.rollaxis(texture_pc_uv, 0, 3)

        # Concatenate to create the data term steepest descent
        J = np.hstack((J, dt_db))

        # Reshape to : n_params x (2 * N)
        n_params = J.shape[1]
        J = np.transpose(J, (1, 0, 2)).reshape(n_params, -1)
        return reconstruction_prior_weight * J, n_camera_parameters

    def __str__(self):
        return "Simultaneous Forward Additive"


class WibergForwardAdditive(LucasKanade):
    r"""
    Class for defining Wiberg Forward Additive Morphable Model optimization
    algorithm.
    """

    def run(
        self,
        image,
        initial_mesh,
        camera,
        gt_mesh=None,
        max_iters=20,
        camera_update=False,
        focal_length_update=False,
        reconstruction_weight=1.0,
        shape_prior_weight=1.0,
        texture_prior_weight=1.0,
        landmarks=None,
        landmarks_prior_weight=1.0,
        return_costs=False,
        verbose=True,
    ):
        # Parse landmarks prior options
        if landmarks is None or landmarks_prior_weight is None:
            landmarks_prior_weight = None
            landmarks = None
        lms_points = None
        if landmarks is not None:
            lms_points = landmarks.points[:, [1, 0]]

        # Retrieve camera parameters from the provided camera object.
        # Project provided instance to retrieve shape and texture parameters.
        camera_parameters = camera.as_vector()
        shape_parameters = self.model.shape_model.project(initial_mesh)
        texture_parameters = self.model.project_instance_on_texture_model(initial_mesh)

        # Reconstruct provided instance
        instance = self.model.instance(
            shape_weights=shape_parameters, texture_weights=texture_parameters
        )

        # Compute input image gradient
        grad_x, grad_y = self.gradient(image)

        # Initialize lists
        shape_parameters_per_iter = [shape_parameters]
        texture_parameters_per_iter = [texture_parameters]
        camera_per_iter = [camera]
        instance_per_iter = [instance.rescale_texture(0.0, 1.0)]
        costs = None
        if return_costs:
            costs = []

        # Initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Main loop
        while k < max_iters and eps > self.eps:
            if verbose:
                print_dynamic("{}/{}".format(k + 1, max_iters))
            # Apply camera projection on current instance
            instance_in_image = camera.apply(instance)

            # Compute indices locations for sampling
            (vertex_indices, bcoords, tri_indices, yx) = self.visible_sample_points(
                instance_in_image, image.shape
            )

            # Warp the mesh with the view matrix (rotation + translation)
            instance_w = camera.view_transform.apply(instance.points)

            # Sample all the terms from the model part at the sample locations
            warped_uv = self.sample(instance_w, bcoords, vertex_indices)
            texture_pc_uv = self.model.sample_texture_model(bcoords, tri_indices)
            texture_pc_uv = texture_pc_uv.reshape((-1, self.m))
            m_texture_uv = (
                self.model.instance().sample_texture_with_barycentric_coordinates(
                    bcoords, tri_indices
                )
            )
            shape_pc_uv = self.sample(self.shape_pc, bcoords, vertex_indices)
            # Reshape shape basis after sampling
            shape_pc_uv = shape_pc_uv.reshape([self.n_samples, 3, -1])

            # Sample all the terms from the image part at the sample locations
            img_uv = image.sample(yx)
            grad_x_uv = grad_x.sample(yx)
            grad_y_uv = grad_y.sample(yx)

            # Compute error
            img_error_uv = (img_uv - m_texture_uv.T).ravel()

            # Compute Jacobian, SD and Hessian of data term
            if reconstruction_weight is not None:
                sd, n_camera_parameters = self.J_data(
                    camera,
                    warped_uv,
                    shape_pc_uv,
                    texture_pc_uv,
                    grad_x_uv,
                    grad_y_uv,
                    camera_update,
                    focal_length_update,
                    reconstruction_weight,
                )
                hessian = sd.dot(sd.T)
                sd_error = sd.dot(img_error_uv)
            else:
                n_camera_parameters = 0
                if camera_update:
                    if focal_length_update:
                        n_camera_parameters = camera.n_parameters - 1
                    else:
                        n_camera_parameters = camera.n_parameters - 2
                hessian = np.zeros(
                    (self.n + n_camera_parameters, self.n + n_camera_parameters)
                )
                sd_error = np.zeros(self.n + n_camera_parameters)

            # Compute Jacobian, update SD and Hessian wrt shape prior
            if shape_prior_weight is not None:
                sd_shape = shape_prior_weight * self.J_shape_prior
                hessian[: self.n, : self.n] += np.diag(sd_shape)
                sd_error[: self.n] += sd_shape * shape_parameters

            # Compute Jacobian, update SD and Hessian wrt landmarks prior
            lms_error = None
            if landmarks_prior_weight is not None:
                # Get projected instance on landmarks and error term
                warped_lms = instance_in_image.points[self.model.model_landmarks_index]
                lms_error = (warped_lms[:, [1, 0]] - lms_points).T.ravel()
                warped_view_lms = instance_w[self.model.model_landmarks_index]

                # Jacobian and Hessian wrt shape parameters
                sd_lms, n_camera_parameters = self.J_lms(
                    camera,
                    warped_view_lms,
                    self.shape_pc_lms,
                    camera_update,
                    focal_length_update,
                )
                idx = self.n + n_camera_parameters
                hessian[:idx, :idx] += landmarks_prior_weight * sd_lms.dot(sd_lms.T)
                sd_error[:idx] = landmarks_prior_weight * sd_lms.dot(lms_error)

            if return_costs:
                costs.append(
                    self.compute_cost(
                        img_error_uv,
                        lms_error,
                        shape_parameters,
                        texture_parameters,
                        shape_prior_weight,
                        texture_prior_weight,
                        landmarks_prior_weight,
                    )
                )

            # Solve to find the increment of parameters
            d_shape, d_camera = self.solve(
                hessian, sd_error, camera_update, focal_length_update
            )

            # Update parameters
            shape_parameters += d_shape
            if camera_update:
                camera_parameters = camera_parameters_update(
                    camera_parameters, d_camera
                )
                camera = camera.from_vector(camera_parameters)

            # Generate the updated instance
            instance = self.model.instance(
                shape_weights=shape_parameters, texture_weights=texture_parameters
            )

            # Update lists
            shape_parameters_per_iter.append(shape_parameters)
            texture_parameters_per_iter.append(texture_parameters)
            camera_per_iter.append(camera)
            instance_per_iter.append(instance.rescale_texture(0.0, 1.0))

            # Increase iteration counter
            k += 1

            # shape_parameters, texture_parameters, camera_parameters = yield \
            #     shape_parameters, texture_parameters, camera_parameters

        return MMAlgorithmResult(
            shape_parameters=shape_parameters_per_iter,
            texture_parameters=texture_parameters_per_iter,
            meshes=instance_per_iter,
            camera_transforms=camera_per_iter,
            image=image,
            initial_mesh=initial_mesh.rescale_texture(0.0, 1.0),
            initial_camera_transform=camera_per_iter[0],
            gt_mesh=gt_mesh,
            costs=costs,
        )

    def project_out(self, J, U):
        tmp = J.dot(U)
        return J - tmp.dot(U.T)

    def compute_hessian(self, sd):
        n_params = sd.shape[1]
        sd = np.transpose(sd, (1, 0, 2)).reshape(n_params, -1)
        return sd.dot(sd.T)

    def J_data(
        self,
        camera,
        warped_uv,
        shape_pc_uv,
        texture_pc_uv,
        grad_x_uv,
        grad_y_uv,
        camera_update,
        focal_length_update,
        reconstruction_prior_weight,
    ):
        # Compute derivative of camera wrt shape and camera parameters
        dp_da_dr = d_camera_d_shape_parameters(camera, warped_uv, shape_pc_uv)
        n_camera_parameters = 0
        if camera_update:
            dp_dr = d_camera_d_camera_parameters(
                camera, warped_uv, with_focal_length=focal_length_update
            )
            dp_da_dr = np.hstack((dp_da_dr, dp_dr))
            n_camera_parameters = dp_dr.shape[1]

        # Multiply image gradient with camera derivative
        permuted_grad_x = np.transpose(grad_x_uv[..., None], (0, 2, 1))
        permuted_grad_y = np.transpose(grad_y_uv[..., None], (0, 2, 1))
        J = permuted_grad_x * dp_da_dr[0] + permuted_grad_y * dp_da_dr[1]

        # Project-out
        n_params = J.shape[1]
        J = np.transpose(J, (1, 0, 2)).reshape(n_params, -1)
        PJ = self.project_out(J, texture_pc_uv)

        # Concatenate to create the data term steepest descent
        return reconstruction_prior_weight * PJ, n_camera_parameters

    def solve(self, hessian, sd_error, camera_update, focal_length_update):
        # Solve
        ds = -np.linalg.solve(hessian, sd_error)

        # Get shape parameters increment
        d_shape = ds[: self.n]

        # Get camera parameters increment
        if camera_update:
            # Keep the rest
            ds = ds[self.n :]

            # If focal length is not updated, then set its increment to zero
            if not focal_length_update:
                ds = np.insert(ds, 0, [0.0])

            # Set increment of the 1st quaternion to one
            ds = np.insert(ds, 1, [1.0])

            # Get camera parameters update
            d_camera = ds
        else:
            d_camera = None

        return d_shape, d_camera

    def _precompute(self):
        # call super method
        super(WibergForwardAdditive, self)._precompute()
        self.texture_T = self.model.texture_model.components.T

    def __str__(self):
        return "Wiberg Forward Additive"


def quaternion_multiply(current_q, increment_q):
    # Make sure that the q increment has unit norm
    increment_q /= np.linalg.norm(increment_q)
    # Update
    w0, x0, y0, z0 = current_q
    w1, x1, y1, z1 = increment_q
    return np.array(
        [
            -x1 * x0 - y1 * y0 - z1 * z0 + w1 * w0,
            x1 * w0 + y1 * z0 - z1 * y0 + w1 * x0,
            -x1 * z0 + y1 * w0 + z1 * x0 + w1 * y0,
            x1 * y0 - y1 * x0 + z1 * w0 + w1 * z0,
        ],
        dtype=np.float64,
    )


def camera_parameters_update(c, dc):
    # Add for focal length and translation parameters, but multiply for
    # quaternions
    new = c + dc
    new[1:5] = quaternion_multiply(c[1:5], dc[1:5])
    return new



from .viewmayavi import (
    MayaviTriMeshViewer3d,
    MayaviPointGraphViewer3d,
    MayaviTexturedTriMeshViewer3d,
    MayaviLandmarkViewer3d,
    MayaviVectorViewer3d,
    MayaviColouredTriMeshViewer3d,
)

PointGraphViewer3d = MayaviPointGraphViewer3d
TriMeshViewer3d = MayaviTriMeshViewer3d
TexturedTriMeshViewer3d = MayaviTexturedTriMeshViewer3d
ColouredTriMeshViewer3d = MayaviColouredTriMeshViewer3d
LandmarkViewer3d = MayaviLandmarkViewer3d
VectorViewer3d = MayaviVectorViewer3d



import numpy as np
from menpo.transform.piecewiseaffine.base import barycentric_vectors
from menpo.image import BooleanImage, MaskedImage


def _pixels_to_check_python(start, end, n_pixels):
    pixel_locations = np.empty((n_pixels, 2), dtype=int)
    tri_indices = np.empty(n_pixels, dtype=int)

    n = 0
    for i, ((s_x, s_y), (e_x, e_y)) in enumerate(zip(start, end)):
        for x in range(s_x, e_x):
            for y in range(s_y, e_y):
                pixel_locations[n] = (x, y)
                tri_indices[n] = i
                n += 1

    return pixel_locations, tri_indices


try:
    from .tripixel import pixels_to_check
except ImportError:
    print("Falling back to CPU pixel checking")
    pixels_to_check = _pixels_to_check_python


def pixel_locations_and_tri_indices(mesh):
    vertex_trilist = mesh.points[mesh.trilist]
    start = np.floor(vertex_trilist.min(axis=1)[:, :2])
    end = np.ceil(vertex_trilist.max(axis=1)[:, :2])
    start = start.astype(int)
    end = end.astype(int)
    n_sites = np.product((end - start), axis=1).sum()
    return pixels_to_check(start, end, n_sites)


def alpha_beta(i, ij, ik, points):
    ip = points - i
    dot_jj = np.einsum("dt, dt -> t", ij, ij)
    dot_kk = np.einsum("dt, dt -> t", ik, ik)
    dot_jk = np.einsum("dt, dt -> t", ij, ik)
    dot_pj = np.einsum("dt, dt -> t", ip, ij)
    dot_pk = np.einsum("dt, dt -> t", ip, ik)

    d = 1.0 / (dot_jj * dot_kk - dot_jk * dot_jk)
    alpha = (dot_kk * dot_pj - dot_jk * dot_pk) * d
    beta = (dot_jj * dot_pk - dot_jk * dot_pj) * d
    return alpha, beta


def xy_bcoords(mesh, tri_indices, pixel_locations):
    i, ij, ik = barycentric_vectors(mesh.points[:, :2], mesh.trilist)
    i = i[:, tri_indices]
    ij = ij[:, tri_indices]
    ik = ik[:, tri_indices]
    a, b = alpha_beta(i, ij, ik, pixel_locations.T)
    c = 1 - a - b
    bcoords = np.array([c, a, b]).T
    return bcoords


def tri_containment(bcoords):
    alpha, beta, _ = bcoords.T
    return np.logical_and(np.logical_and(alpha >= 0, beta >= 0), alpha + beta <= 1)


def z_values_for_bcoords(mesh, bcoords, tri_indices):
    return mesh.barycentric_coordinate_interpolation(
        mesh.points[:, -1][..., None], bcoords, tri_indices
    )[:, 0]


def pixel_sample_uniform(xy, n_samples):
    chosen_mask = np.random.permutation(np.arange(xy.shape[0]))[:n_samples]
    return xy[chosen_mask]


def unique_locations(xy, width, height):
    mask = np.zeros([width, height], dtype=np.bool)
    mask[xy[:, 0], xy[:, 1]] = True
    return np.vstack(np.nonzero(mask)).T


def location_to_index(xy, width):
    return xy[:, 0] * width + xy[:, 1]


def rasterize_barycentric_coordinates(mesh, image_shape):
    height, width = int(image_shape[0]), int(image_shape[1])
    # 1. Find all pixel-sites that may need to be rendered to
    #    + the triangle that may partake in rendering
    yx, tri_indices = pixel_locations_and_tri_indices(mesh)

    # 2. Limit to only pixel sites in the image
    out_of_bounds = np.logical_or(
        np.any(yx < 0, axis=1), np.any((np.array([height, width]) - yx) <= 0, axis=1)
    )
    in_image = ~out_of_bounds
    yx = yx[in_image]
    tri_indices = tri_indices[in_image]

    # # Optionally limit to subset of pixels
    # if n_random_samples is not None:
    #     # 2. Find the unique pixel sites
    #     xy_u = unique_locations(yx, width, height)
    #
    #     xy_u = pixel_sample_uniform(xy_u, n_random_samples)
    #     to_keep = np.in1d(location_to_index(yx, width),
    #                       location_to_index(xy_u, width))
    #     yx = yx[to_keep]
    #     tri_indices = tri_indices[to_keep]

    bcoords = xy_bcoords(mesh, tri_indices, yx)

    # check the mask based on triangle containment
    in_tri_mask = tri_containment(bcoords)

    # use this mask on the pixels
    yx = yx[in_tri_mask]
    bcoords = bcoords[in_tri_mask]
    tri_indices = tri_indices[in_tri_mask]

    # Find the z values for all pixels and calculate the mask
    z_values = z_values_for_bcoords(mesh, bcoords, tri_indices)

    # argsort z from smallest to biggest - use this to sort all data
    sort = np.argsort(z_values)
    yx = yx[sort]
    bcoords = bcoords[sort]
    tri_indices = tri_indices[sort]

    # make a unique id per-pixel location
    pixel_index = yx[:, 0] * width + yx[:, 1]
    # find the first instance of each pixel site by depth
    _, z_buffer_mask = np.unique(pixel_index, return_index=True)

    # mask the locations one last time
    yx = yx[z_buffer_mask]
    bcoords = bcoords[z_buffer_mask]
    tri_indices = tri_indices[z_buffer_mask]
    return yx, bcoords, tri_indices


def rasterize_barycentric_coordinate_images(mesh, image_shape):
    h, w = image_shape
    yx, bcoords, tri_indices = rasterize_barycentric_coordinates(mesh, image_shape)

    tri_indices_img = np.zeros((1, h, w), dtype=int)
    bcoords_img = np.zeros((3, h, w))
    mask = np.zeros((h, w), dtype=np.bool)
    mask[yx[:, 0], yx[:, 1]] = True
    tri_indices_img[:, yx[:, 0], yx[:, 1]] = tri_indices
    bcoords_img[:, yx[:, 0], yx[:, 1]] = bcoords.T

    mask = BooleanImage(mask)
    return (
        MaskedImage(bcoords_img, mask=mask.copy(), copy=False),
        MaskedImage(tri_indices_img, mask=mask.copy(), copy=False),
    )



#cython: boundscheck=False, wraparound=False, nonecheck=False
cimport numpy as np
import numpy as np


def pixels_to_check(long[:, :] start, long[:, :] end, long n_pixels):

    cdef long [:, :] pixel_locations = np.empty((n_pixels, 2), dtype=long)
    cdef long [:] tri_indices = np.empty(n_pixels, dtype=long)

    cdef long n_tris = start.shape[0]
    cdef long i, j, k, sx, sy, ex, ey
    cdef long n = 0

    for i in range(n_tris):
        sx = start[i, 0]
        sy = start[i, 1]
        ex = end[i, 0]
        ey = end[i, 1]
        for j in range(sx, ex):
            for k in range(sy, ey):
                pixel_locations[n, 0] = j
                pixel_locations[n, 1] = k
                tri_indices[n] = i
                n += 1

    return np.asarray(pixel_locations), np.asarray(tri_indices)



from pathlib import Path
from glob import glob

SHADERS_DIR = Path(__file__).parents[1] / "shaders"


def test_n_expected_builtin_shaders():
    all_shaders = glob(str(SHADERS_DIR / "**/*.*"), recursive=True)
    assert len(all_shaders) == 4, all_shaders


def test_all_per_vertex_shaders_exist():
    per_vertex_shaders = SHADERS_DIR / "per_vertex"
    assert [p.name for p in sorted(per_vertex_shaders.glob("*"))] == [
        "passthrough.frag",
        "passthrough.vert",
    ]


def test_all_texture_shaders_exist():
    per_vertex_shaders = SHADERS_DIR / "texture"
    assert [p.name for p in sorted(per_vertex_shaders.glob("*"))] == [
        "passthrough.frag",
        "passthrough.vert",
    ]
